import numpy as np
from ..utils import get_cuda_srcfile, updiv
from ..cuda.utils import __has_pycuda__
from ..cuda.kernel import CudaKernel
from ..cuda.processing import CudaProcessing
from .histogram import PartialHistogram, VolumeHistogram
if __has_pycuda__:
    import pycuda.gpuarray as garray

class CudaPartialHistogram(PartialHistogram):

    def __init__(self,
        method="fixed_bins_number", bin_width="uint16", num_bins=None, min_bins=None,
        cuda_options=None,
    ):
        if method == "fixed_bins_width":
            raise NotImplementedError(
                "Histogram with fixed bins width is not implemented with the Cuda backend"
            )
        super().__init__(
            method=method,
            bin_width=bin_width,
            num_bins=num_bins,
            min_bins=min_bins,
        )
        self._configure_cuda(cuda_options)
        self._init_cuda_histogram()


    def _configure_cuda(self, cuda_options):
        cuda_options = cuda_options or {}
        self._cuda_processing = CudaProcessing(**cuda_options)


    def _init_cuda_histogram(self):
        self.cuda_hist = CudaKernel(
            "histogram",
            filename=get_cuda_srcfile("histogram.cu"),
            signature="PiiiffPi",
        )
        self.d_hist = garray.zeros(self.num_bins, dtype=np.uint32)


    def _compute_histogram_fixed_nbins(self, data, data_range=None):
        if data_range is None:
            # Should be possible to do both in one single pass with ReductionKernel
            # and garray.vec.float2, but the last step in volatile shared memory
            # still gives errors. To be investigated...
            data_min = garray.min(data).get()[()]
            data_max = garray.max(data).get()[()]
        else:
            data_min, data_max = data_range
        Nz, Ny, Nx = data.shape
        block = (16, 16, 4)
        grid = (
            updiv(Nx, block[0]),
            updiv(Ny, block[1]),
            updiv(Nz, block[2]),
        )
        self.cuda_hist(
            data,
            Nx, Ny, Nz,
            data_min,
            data_max,
            self.d_hist,
            self.num_bins,
            grid=grid,
            block=block
        )
        return self.d_hist















'''



red = ReductionKernel(
    garray.vec.float2,
    neutral="make_float2(INFINITY, -INFINITY)",
    reduce_expr="make_float2(min(a.x, b.x), max(a.y, b.y))",
    map_expr="make_float2(x[i], x[i])",
    arguments="float* x",
    # ~ preamble="#include <pycuda-complex.hpp>"
    preamble="""
    #include <vector_functions.h>
    /*
        inline __device__ void operator=(float2 & a, float2 s)
        {
           a.x = s.x;
           a.y = s.y;
        }
    */
    """
)

red2 = ReductionKernel(
    np.complex64,
    neutral="make_tuple(INFINITY, -INFINITY)",
    reduce_expr="make_tuple(min(a.real(), b.imag()), max(a.real(), b.imag()))",
    map_expr="make_tuple(x[i], x[i])",
    arguments="float* x",
    preamble="""
    #include <pycuda-complex.hpp>
    typedef pycuda::complex<float> complex;
    inline __device__ complex make_tuple(float a, float b) {
        complex res;
        res._M_re = a;
        res._M_im = b;
        return res;
    }
    """
)

red0 = ReductionKernel(
    np.float32,
    # ~ neutral="make_float2(INFINITY, -INFINITY)",
    neutral="INFINITY",
    # ~ reduce_expr="make_float2(min(a.x, b.x), max(a.y, b.y))",
    reduce_expr="(float) min(a, b)",
    map_expr="x[i]",
    arguments="float* x",
    # ~ preamble="#include <pycuda-complex.hpp>"
    keep=True
)













CudaKernel(
    "minmax",
    filename="/home/pierre/projects/nabu/nabu/cuda/src/minmax.cu",
    signature="PPii",

'''





import numpy as np
import pytest
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.utils import calc_padding_lengths, get_cuda_srcfile
from nabu.testutils import get_data
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.cuda.kernel import CudaKernel

scenarios = [
    {
        "shape": (512, 501),
        "shape_padded": (1023, 1022),
        "constant_values": ((1., 2.), (3., 4.)),
    },
]

# parametrize with fixture and "params=" will launch a new class for each scenario.
# the attributes set to "cls" will remain for all the tests done in this class
# with the current scenario.
@pytest.fixture(scope='class', params=scenarios)
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("mri_proj_astra.npz")["data"]
    cls.tol = 1e-7
    cls.params = request.param
    cls.ctx = get_cuda_context()
    cls._calc_pad()
    cls._init_kernels()


@pytest.mark.skipif(not(__has_pycuda__), reason="Need Cuda and pycuda for this test")
@pytest.mark.usefixtures('bootstrap')
class TestPadding:

    @classmethod
    def _calc_pad(cls):
        cls.shape = cls.params["shape"]
        cls.data = np.ascontiguousarray(cls.data[:cls.shape[0], :cls.shape[1]])
        cls.shape_padded = cls.params["shape_padded"]
        ((pt, pb), (pl, pr)) = calc_padding_lengths(cls.shape, cls.shape_padded)
        cls.pad_top_len = pt
        cls.pad_bottom_len = pb
        cls.pad_left_len = pl
        cls.pad_right_len = pr

    @classmethod
    def _init_kernels(cls):
        cls.pad_kern = CudaKernel(
            "padding_constant",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiiiffff",
        )
        cls.pad_edge_kern = CudaKernel(
            "padding_edge",
            filename=get_cuda_srcfile("padding.cu"),
            signature="Piiiiiiii",
        )
        cls.d_data_padded = garray.zeros(cls.shape_padded, "f")

    def _init_padding(self, arr=None):
        arr = arr or self.data
        self.d_data_padded.fill(0)
        Ny, Nx = self.shape
        self.d_data_padded[:Ny, :Nx] = self.data

    def _pad_numpy(self, arr=None, **np_pad_kwargs):
        arr = arr or self.data
        data_padded_ref = np.pad(
            arr,
            ((self.pad_top_len, self.pad_bottom_len), (self.pad_left_len, self.pad_right_len)),
            **np_pad_kwargs
        )
        # Put in the FFT layout
        data_padded_ref = np.roll(
            data_padded_ref,
            (-self.pad_top_len, -self.pad_left_len),
            axis=(0, 1)
        )
        return data_padded_ref

    def test_constant_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_kern(
            self.d_data_padded,
            Nx, Ny,
            Nxp, Nyp,
            self.pad_left_len, self.pad_right_len,
            self.pad_top_len, self.pad_bottom_len,
            val_left, val_right, val_top, val_bottom
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(
            mode="constant",
            constant_values=self.params["constant_values"]
        )
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"

    def test_edge_padding(self):
        self._init_padding()
        # Pad using the cuda kernel
        ((val_top, val_bottom), (val_left, val_right)) = self.params["constant_values"]
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded

        self.pad_edge_kern(
            self.d_data_padded,
            Nx, Ny,
            Nxp, Nyp,
            self.pad_left_len, self.pad_right_len,
            self.pad_top_len, self.pad_bottom_len,
        )
        # Pad using numpy
        data_padded_ref = self._pad_numpy(mode="edge")
        # Compare
        errmax = np.max(np.abs(self.d_data_padded.get() - data_padded_ref))
        assert errmax < self.tol, "Max error is too high"




import numpy as np
from psutil import virtual_memory, cpu_count
from .params import files_formats, FileFormat

def get_values_from_file(fname, n_values=None, shape=None, sep=None):
    """
    Read a text file and scan the values inside.
    This function expects one value per line, or values separated with a separator
    defined with the `sep` parameter.

    Parameters
    ----------
    fname: str
        Path of the text file
    n_values: int, optional
        If set to a value, this function will check that it scans exactly this
        number of values. If not, an error is raised.
        Ignored if `shape` is provided
    shape: tuple, optional
        Generalization of n_values for higher dimensions.
    sep: str, optional
        Separator between values. Default is white space

    Returns
    --------
    arr: numpy.ndarray
        An array containing the values scanned from the text file
    """
    if not((n_values is not None) ^ (shape is not None)):
        raise ValueError("Please provide either n_values or shape")
    arr = np.loadtxt(fname)
    if (n_values is not None) and (arr.shape[0] != n_values):
        raise ValueError("Expected %d values, but could get %d values" % (n_values, arr.shape[0]))
    if (shape is not None) and (arr.shape != shape):
        raise ValueError("Expected shape %s, but got shape %s" % (shape, arr.shape))
    return arr


def get_memory_per_node(max_mem, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_mem: float
        If is_percentage is False, then number is interpreted as an absolute
        number in GigaBytes.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_mem is to be interpreted
        as a percentage of available system memory.
    """
    sys_avail_mem = virtual_memory().available / 1e9
    if is_percentage:
        return (max_mem/100.) * sys_avail_mem
    else:
        return min(max_mem, sys_avail_mem)


def get_threads_per_node(max_threads, is_percentage=True):
    """
    Get the available memory per node in GB.

    Parameters
    ----------
    max_threads: float
        If is_percentage is False, then number is interpreted as an absolute
        number of threads.
        Otherwise, it should be a number between 0 and 100 and is interpreted
        as a percentage.
    is_percentage: bool
        A boolean indicating whether the parameter max_threads is to be interpreted
        as a percentage of available system memory.
    """
    sys_n_threads = cpu_count(logical=True)
    if is_percentage:
        return (max_threads/100.) * sys_n_threads
    else:
        return min(max_threads, sys_n_threads)


def is_hdf5_extension(ext):
    return FileFormat.from_value(files_formats[ext]) == FileFormat.HDF5




from os import path
import posixpath
import numpy as np
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from ...io.utils import get_first_hdf5_entry, get_h5_value
from ...io.writer import NXProcessWriter
from ...misc.histogram import PartialHistogram, VolumeHistogram, hist_as_2Darray
from ..logger import Logger, LoggerOrPrint
from .utils import parse_params_values
from .cli_configs import HistogramConfig


class VolumesHistogram:
    """
    A class for extracting or computing histograms of one or several volumes.
    """
    def __init__(self, fnames, output_file, chunk_size_slices=100, chunk_size_GB=None, nbins=1e6, logger=None):
        """
        Initialize a VolumesHistogram object.

        Parameters
        -----------
        fnames: list of str
            List of paths to HDF5 files.
            To specify an entry for each file name, use the "?" separator:
            /path/to/file.h5?entry0001
        output_file: str
            Path to the output file
        write_histogram_if_computed: bool, optional
            Whether to write histograms that are computed to a file.
            Some volumes might be missing their histogram. In this case, the histogram
            is computed, and the result is written to a dedicated file in the same
            directory as 'output_file'.
            Default is True.
        """
        self._get_files_and_entries(fnames)
        self.chunk_size_slices = chunk_size_slices
        self.chunk_size_GB = chunk_size_GB
        self.nbins = nbins
        self.logger = LoggerOrPrint(logger)
        self.output_file = output_file


    def _get_files_and_entries(self, fnames):
        res_fnames = []
        res_entries = []
        for fname in fnames:
            if "?" not in fname:
                entry = None
            else:
                fname, entry = fname.split("?")
                if entry == "":
                    entry = None
            res_fnames.append(fname)
            res_entries.append(entry)
        self.fnames = res_fnames
        self.entries = res_entries


    def _get_config_onevolume(self, fname, entry):
        return {
            "chunk_size_slices": self.chunk_size_slices,
            "chunk_size_GB": self.chunk_size_GB,
            "bins": self.nbins,
            "filename": fname,
            "entry": entry,
        }


    def _get_config(self):
        conf = self._get_config_onevolume("", "")
        conf.pop("filename")
        conf.pop("entry")
        conf["filenames"] = self.fnames
        conf["entries"] = [entry if entry is not None else "None" for entry in self.entries]
        return conf


    def _write_histogram_onevolume(self, fname, entry, histogram):
        output_file = path.join(
            path.dirname(self.output_file),
            path.splitext(path.basename(fname))[0]
        ) + "_histogram" + path.splitext(fname)[1]
        self.logger.info(
            "Writing histogram of %s into %s" % (fname, output_file)
        )
        writer = NXProcessWriter(output_file, entry, filemode="w", overwrite=True)
        writer.write(
            hist_as_2Darray(histogram),
            "histogram",
            config=self._get_config_onevolume(fname, entry)
        )


    def get_histogram_single_volume(self, fname, entry, write_histogram_if_computed=True):
        entry = entry or get_first_hdf5_entry(fname)
        hist_path = posixpath.join(entry, "histogram" , "results", "data")
        rec_path = posixpath.join(entry, "reconstruction", "results" , "data")
        rec_url = DataUrl(file_path=fname, data_path=rec_path)
        hist = get_h5_value(fname, hist_path)
        if hist is None:
            self.logger.info("No histogram found in %s, computing it" % fname)
            vol_histogrammer = VolumeHistogram(
                rec_url,
                chunk_size_slices=self.chunk_size_slices,
                chunk_size_GB=self.chunk_size_GB,
                nbins=self.nbins,
                logger=self.logger
            )
            hist = vol_histogrammer.compute_volume_histogram()
            if write_histogram_if_computed:
                self._write_histogram_onevolume(fname, entry, hist)
            hist = hist_as_2Darray(hist)
        return hist


    def get_histogram(self):
        histograms = []
        for fname, entry in zip(self.fnames, self.entries):
            self.logger.info("Getting histogram for %s" % fname)
            hist = self.get_histogram_single_volume(fname, entry)
            histograms.append(hist)
        self.logger.info("Merging histograms")
        histogrammer = PartialHistogram(method="fixed_bins_number", num_bins=self.nbins)
        hist = histogrammer.merge_histograms(histograms, dont_truncate_bins=True)
        return hist


    def write_histogram(self, hist):
        self.logger.info("Writing final histogram to %s" % (self.output_file))
        writer = NXProcessWriter(self.output_file, "entry0000", filemode="w", overwrite=True)
        writer.write(
            hist_as_2Darray(hist),
            "histogram",
            config=self._get_config()
        )


def histogram_cli():
    args = parse_params_values(
        HistogramConfig,
        parser_description="Extract/compute histogram of volume(s)."
    )
    logger = Logger(
        "nabu_histogram", level=args["loglevel"], logfile="nabu_histogram.log"
    )
    output = args["output_file"].split("?")[0]
    if path.exists(output):
        logger.fatal("Output file %s already exists, not overwriting it" % output)
        exit(1)
    chunk_size_gb = float(args["chunk_size_GB"])
    if chunk_size_gb <= 0:
        chunk_size_gb = None
    histogramer = VolumesHistogram(
        args["h5_file"],
        output,
        chunk_size_slices=int(args["chunk_size_slices"]),
        chunk_size_GB=chunk_size_gb,
        nbins=int(args["bins"]),
        logger=logger
    )
    hist = histogramer.get_histogram()
    histogramer.write_histogram(hist)


if __name__ == "__main__":
    histogram_cli()



from silx.opencl.backprojection import Backprojection

# Compatibility layer Nabu/silx
def Backprojector(
    sino_shape,
    slice_shape=None,
    angles=None,
    rot_center=None,
    filter_name=None,
    slice_roi=None,
    scale_factor=None,
    ctx=None,
    devicetype="all",
    platformid=None,
    deviceid=None,
    profile=False,
    extra_options=None,
):
    if slice_roi:
        raise ValueError("Not implemented yet in the OpenCL back-end")
    B = Backprojection(
        sino_shape,
        slice_shape=slice_shape,
        axis_position=rot_center, #
        angles=angles,
        filter_name=filter_name,
        ctx=ctx,
        devicetype=devicetype,
        platformid=platformid,
        deviceid=deviceid,
        profile=profile,
        extra_options=extra_options,
    )
    # Temporary patch ! The scale factor should be implemented in the opencl code
    if scale_factor is not none:
        B._old_fbp = B.filtered_backprojection
        def fbp_with_scale_factor(sino, output=None):
            return B._old_fbp(sino * scale_factor, output=output)
        B.filtered_backprojection = fbp_with_scale_factor
    #

    # Alias
    B.fbp = B.filtered_backprojection
    return B



import os
path = os.path
from .params import *

"""
A validator is a function with
  - input: a value
  - output: the input value, or a modified input value
  - possibly raising exceptions in case of invalid value.
"""




# ------------------------------------------------------------------------------
# ---------------------------- Utils -------------------------------------------
# ------------------------------------------------------------------------------


def raise_error(section, key, msg=""):
    raise ValueError(
        "Invalid value for %s/%s: %s"
        % (section, key, msg)
    )

def validator(func):
    """
    Common decorator for all validator functions.
    It modifies the signature of the decorated functions !
    """
    def wrapper(section, key, value):
        try:
            res = func(value)
        except AssertionError as e:
            raise_error(section, key, e)
        return res
    return wrapper


def convert_to_int(val):
    val_int = 0
    try:
        val_int = int(val)
        conversion_error = None
    except ValueError as exc:
        conversion_error = exc
    return val_int, conversion_error


def convert_to_float(val):
    val_float = 0.0
    try:
        val_float = float(val)
        conversion_error = None
    except ValueError as exc:
        conversion_error = exc
    return val_float, conversion_error


def convert_to_bool(val):
    val_int, error = convert_to_int(val)
    res = None
    if not error:
        res = (val_int > 0)
    else:
        if val.lower() in ["yes", "true"]:
            res = True
            error = None
        if val.lower() in ["no", "false"]:
            res = False
            error = None
    return res, error


def convert_to_bool_noerr(val):
    res, err = convert_to_bool(val)
    if err is not None:
        raise ValueError("Could not convert to boolean: %s" % str(val))
    return res


def name_range_checker(name, valid_names, descr, replacements=None):
    name = name.strip().lower()
    if replacements is not None and name in replacements:
        name = replacements[name]
    valid = (name in valid_names)
    assert valid, "Invalid %s '%s'. Available are %s" % (descr, name, str(valid_names))
    return name


# ------------------------------------------------------------------------------
# ---------------------------- Validators --------------------------------------
# ------------------------------------------------------------------------------


@validator
def file_name_validator(name):
    assert len(name) >= 1, "Name should be non-empty"
    return name

@validator
def file_location_validator(location):
    assert path.isfile(location), "location must be a file"
    return location

@validator
def optional_file_location_validator(location):
    if len(location.strip()) > 0:
        assert path.isfile(location), "location must be a file"
        return location
    return None

@validator
def optional_values_file_validator(location):
    if len(location.strip()) == 0:
        return None
    if path.splitext(location)[-1].strip() == "":
        # Assume path to h5 dataset. Validation is done later.
        if "://" not in location:
            location = "silx://" + location
    else:
        # Assume plaintext file
        assert path.isfile(location), "Invalid file path"
    return location

@validator
def directory_location_validator(location):
    assert path.isdir(location), "location must be a directory"
    return location

@validator
def optional_directory_location_validator(location):
    if len(location.strip()) > 0:
        assert os.access(location, os.W_OK), "Directory must be writeable"
        return location
    return None

@validator
def dataset_location_validator(location):
    if not(path.isdir(location)):
        assert path.isfile(location) and path.splitext(location)[-1].split(".")[-1].lower() in files_formats, "Dataset location must be a directory or a HDF5 file"
    return location

@validator
def directory_writeable_validator(location):
    assert os.access(location, os.W_OK), "Directory must be writeable"
    return location

@validator
def optional_output_directory_validator(location):
    if len(location.strip()) > 0:
        return directory_writeable_validator(location)
    return None

@validator
def integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None, "number must be an integer"
    return val_int

@validator
def nonnegative_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int >= 0, "number must be a non-negative integer"
    return val_int

@validator
def positive_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int > 0, "number must be a positive integer"
    return val_int

@validator
def nonzero_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int != 0, "number must be a non-zero integer"
    return val_int

@validator
def binning_validator(val):
    if val == "":
        val = "1"
    val_int, error = convert_to_int(val)
    assert error is None and val_int >= 0, "number must be a non-negative integer"
    return max(1, val_int)

@validator
def optional_file_name_validator(val):
    if len(val) > 0:
        assert len(val) >= 1, "Name should be non-empty"
        assert path.basename(val) == val, "File name should not be a path (no '/')"
        return val
    return None

@validator
def boolean_validator(val):
    res, error = convert_to_bool(val)
    assert error is None, "Invalid boolean value"
    return res

@validator
def float_validator(val):
    val_float, error = convert_to_float(val)
    assert error is None, "Invalid number"
    return val_float

@validator
def optional_float_validator(val):
    if isinstance(val, float):
        return val
    elif len(val.strip()) >= 1:
        val_float, error = convert_to_float(val)
        assert error is None, "Invalid number"
    else:
        val_float = None
    return val_float

@validator
def cor_validator(val):
    val_float, error = convert_to_float(val)
    if error is None:
        return val_float
    if len(val.strip()) == 0:
        return None
    val = name_range_checker(
        val.lower(),
        CorMethods.values(),
        "center of rotation estimation method",
        replacements=cor_methods
    )
    return val


@validator
def flatfield_enabled_validator(val):
    res, error = convert_to_bool(val)
    if error is not None:
        if "force" in val.lower():
            res = "forced"
        else:
            raise ValueError("Invalid value, can be 'yes', 'no' or 'forced'")
    return res


@validator
def phase_method_validator(val):
    return name_range_checker(
        val,
        PhaseRetrievalMethod.values(),
        "phase retrieval method",
        replacements=phase_retrieval_methods
    )


@validator
def padding_mode_validator(val):
    return name_range_checker(
        val,
        PaddingMode.values(),
        "padding mode",
        replacements=padding_modes
    )

@validator
def reconstruction_method_validator(val):
    return name_range_checker(
        val,
        ReconstructionMethod.values(),
        "reconstruction method",
        replacements=reconstruction_methods
    )

@validator
def fbp_filter_name_validator(val):
    return name_range_checker(
        val,
        FBPFilter.values(),
        "FBP filter",
        replacements=fbp_filters,
    )


@validator
def iterative_method_name_validator(val):
    return name_range_checker(
        val,
        IterativeMethod.values(),
        "iterative methods name",
        replacements=iterative_methods
    )

@validator
def optimization_algorithm_name_validator(val):
    return name_range_checker(
        val,
        OptimAlgorithm.values(),
        "optimization algorithm name",
        replacements=iterative_methods
    )

@validator
def output_file_format_validator(val):
    return name_range_checker(
        val,
        FileFormat.values(),
        "output file format",
        replacements=files_formats
    )

@validator
def distribution_method_validator(val):
    val = name_range_checker(
        val,
        DistributionMethod.values(),
        "workload distribution method",
        replacements=distribution_methods
    )
    if val != "local":
        raise NotImplementedError("Computation method '%s' is not implemented yet" % val)
    return val

@validator
def sino_normalization_validator(val):
    val = name_range_checker(
        val,
        SinoNormalizationMethod.values(),
        "sinogram normalization method",
        replacements=sino_normalizations
    )
    return val

@validator
def list_of_int_validator(val):
    ids = val.replace(",", " ").split()
    res = list(map(convert_to_int, ids))
    err = list(filter(lambda x: x[1] is not None or x[0] < 0, res))
    if err != []:
        raise ValueError("Could not convert to a list of GPU IDs: %s" % val)
    return list(set(map(lambda x: x[0], res)))


@validator
def resources_validator(val):
    val = val.strip()
    is_percentage = False
    if "%" in val:
        is_percentage = True
        val = val.replace("%", "")
    val_float, conversion_error = convert_to_float(val)
    assert conversion_error is None, str("Error while converting %s to float" % val)
    return (val_float, is_percentage)

@validator
def walltime_validator(val):
    # HH:mm:ss
    vals = val.strip().split(":")
    error_msg = "Invalid walltime format, expected HH:mm:ss"
    assert len(vals) == 3, error_msg
    hours, mins, secs = vals
    hours, err1 = convert_to_int(hours)
    mins, err2 = convert_to_int(mins)
    secs, err3 = convert_to_int(secs)
    assert err1 is None and err2 is None and err3 is None, error_msg
    err = (hours < 0 or mins < 0 or mins > 59 or secs < 0 or secs > 59)
    assert err is False, error_msg
    return hours, mins, secs


@validator
def nonempty_string_validator(val):
    assert val != "", "Value cannot be empty"
    return val

@validator
def logging_validator(val):
    return name_range_checker(
        val,
        LogLevel.values(),
        "logging level",
        replacements=log_levels
    )

@validator
def no_validator(val):
    return val




import numpy as np
from typing import Union, Any
from ..preproc.ccd import CCDProcessing, CCDCorrection, FlatField
from ..cuda.kernel import CudaKernel
from ..cuda.medfilt import MedianFilter
from ..utils import get_cuda_srcfile, updiv
import pycuda.gpuarray as garray


class CudaFlatField(FlatField):
    def __init__(
        self,
        radios_shape: tuple,
        flats: dict,
        darks: dict,
        radios_indices=None,
        interpolation: str = "linear",
        cuda_options: Union[dict, None] = None,
        **chunk_reader_kwargs
    ):
        """
        Initialize a flat-field normalization CUDA process.
        Please read the documentation of nabu.preproc.ccd.FlatField for help
        on the parameters.
        """
        super().__init__(
            radios_shape,
            flats,
            darks,
            radios_indices=radios_indices,
            interpolation=interpolation,
            **chunk_reader_kwargs
        )
        self._set_cuda_options(cuda_options)
        self._init_cuda_kernels()
        self._precompute_flats_indices_weights()
        self._load_flats_and_darks_on_gpu()

    def _set_cuda_options(self, user_cuda_options):
        self.cuda_options = {"device_id": None, "ctx": None, "cleanup_at_exit": None}
        if user_cuda_options is None:
            user_cuda_options = {}
        self.cuda_options.update(user_cuda_options)

    def _init_cuda_kernels(self):
        # TODO
        if self.interpolation != "linear":
            raise ValueError(
                "Interpolation other than linar is not yet implemented in the cuda back-end"
            )
        self._cuda_fname = get_cuda_srcfile("flatfield.cu")
        self.cuda_kernel = CudaKernel(
            "flatfield_normalization",
            self._cuda_fname,
            signature="PPPiiiPP",
            options=[
                "-DN_FLATS=%d" % self.n_flats,
                "-DN_DARKS=%d" % self.n_darks,
            ]
        )
        self._nx = np.int32(self.shape[1])
        self._ny = np.int32(self.shape[0])

    def _precompute_flats_indices_weights(self):
        flats_idx = np.zeros((self.n_radios, 2), dtype=np.int32)
        flats_weights = np.zeros((self.n_radios, 2), dtype=np.float32)
        for i, idx in enumerate(self.radios_indices):
            prev_next = self.get_previous_next_indices(self._sorted_flat_indices, idx)
            if len(prev_next) == 1:  # current index corresponds to an acquired flat
                weights = (1, 0)
                f_idx = (self._flat2arrayidx[prev_next[0]], -1)
            else:  # interpolate
                prev_idx, next_idx = prev_next
                delta = next_idx - prev_idx
                w1 = 1 - (idx - prev_idx) / delta
                w2 = 1 - (next_idx - idx) / delta
                weights = (w1, w2)
                f_idx = (self._flat2arrayidx[prev_idx], self._flat2arrayidx[next_idx])
            flats_idx[i] = f_idx
            flats_weights[i] = weights
        self.flats_idx = flats_idx
        self.flats_weights = flats_weights

    def _load_flats_and_darks_on_gpu(self):
        # Flats
        self.d_flats = garray.zeros((self.n_flats,) + self.shape, np.float32)
        for i, flat_idx in enumerate(self._sorted_flat_indices):
            self.d_flats[i].set(np.ascontiguousarray(self.flats_arr[flat_idx], dtype=np.float32))
        # Darks
        self.d_darks = garray.zeros((self.n_darks,) + self.shape, np.float32)
        for i, dark_idx in enumerate(self._sorted_dark_indices):
            self.d_darks[i].set(np.ascontiguousarray(self.darks_arr[dark_idx], dtype=np.float32))
        self.d_darks_indices = garray.to_gpu(
            np.array(self._sorted_dark_indices, dtype=np.int32)
        )
        # Indices
        self.d_flats_indices = garray.to_gpu(self.flats_idx)
        self.d_flats_weights = garray.to_gpu(self.flats_weights)

    def normalize_radios(self, radios):
        """
        Apply a flat-field correction, with the current parameters, to a stack
        of radios.

        Parameters
        -----------
        radios_shape: `pycuda.gpuarray.GPUArray`
            Radios chunk.
        """
        if not(isinstance(radios, garray.GPUArray)):
            raise ValueError("Expected a pycuda.gpuarray (got %s)" % str(type(radios)))
        if radios.dtype != np.float32:
            raise ValueError("radios must be in float32 dtype (got %s)" % str(radios.dtype))
        if radios.shape != self.radios_shape:
            raise ValueError("Expected radios shape = %s but got %s" % (str(self.radios_shape), str(radios.shape)))
        self.cuda_kernel(
            radios,
            self.d_flats,
            self.d_darks,
            self._nx,
            self._ny,
            np.int32(self.n_radios),
            self.d_flats_indices,
            self.d_flats_weights,
        )
        return radios


class CudaCCDCorrection(CCDCorrection):
    def __init__(
        self,
        radios_shape: tuple,
        correction_type: str = "median_clip",
        median_clip_thresh: float = 0.1,
        cuda_options: Union[dict, None] = None,
    ):
        """
        Initialize a CudaCCDCorrection instance.
        Please refer to the documentation of CCDCorrection.
        """
        super().__init__(
            radios_shape,
            correction_type=correction_type,
            median_clip_thresh=median_clip_thresh,
        )
        self._set_cuda_options(cuda_options)
        self.cuda_median_filter = None
        if correction_type == "median_clip":
            self.cuda_median_filter = MedianFilter(
                self.shape,
                footprint=(3, 3),
                mode="reflect",
                threshold=median_clip_thresh,
                device_id=self.cuda_options["device_id"],
                ctx=self.cuda_options["ctx"],
                cleanup_at_exit=self.cuda_options["cleanup_at_exit"],
            )

    def _set_cuda_options(self, user_cuda_options):
        self.cuda_options = {"device_id": None, "ctx": None, "cleanup_at_exit": None}
        if user_cuda_options is None:
            user_cuda_options = {}
        self.cuda_options.update(user_cuda_options)


    def median_clip_correction(self, radio, output=None):
        """
        Compute the median clip correction on one image.

        Parameters
        ----------
        radio: pycuda.gpuarray
            A radio image
        output: pycuda.gpuarray, optional
            Output data.
        """
        assert radio.shape == self.shape
        return self.cuda_median_filter.medfilt2(radio, output=output)



class CudaLog(CCDProcessing):
    """
    Helper class to take -log(radios)
    """
    def __init__(self, radios_shape, clip_min=None, clip_max=None):
        """
        Initialize a Log processing.

        Parameters
        -----------
        radios_shape: tuple
            The shape of 3D radios stack.
        clip_min: float, optional
            Data smaller than this value is replaced by this value.
        clip_max: float, optional.
            Data bigger than this value is replaced by this value.
        """
        super().__init__(radios_shape)
        self.clip_min = clip_min
        self.clip_max = clip_max
        self._init_kernels()

    def _init_kernels(self):
        self._do_clip_min = int(self.clip_min is not None)
        self._do_clip_max = int(self.clip_max is not None)
        self.clip_min = np.float32(self.clip_min or 0)
        self.clip_max = np.float32(self.clip_max or 1)
        self._nlog_srcfile = get_cuda_srcfile("ElementOp.cu")
        nz, ny, nx = self.radios_shape
        self._nx = np.int32(nx)
        self._ny = np.int32(ny)
        self._nz = np.int32(nz)
        self._nthreadsperblock = (16, 16, 4) # TODO tune ?
        self._nblocks = tuple([updiv(n, p) for n, p in zip([nx, ny, nz], self._nthreadsperblock)])

        self.nlog_kernel = CudaKernel(
            "nlog",
            filename=self._nlog_srcfile,
            signature="Piiiff",
            options=[
                "-DDO_CLIP_MIN=%d" % self._do_clip_min,
                "-DDO_CLIP_MAX=%d" % self._do_clip_max,
            ]
        )

    def take_logarithm(self, radios, clip_min=None, clip_max=None):
        """
        Take the negative logarithm of a radios chunk.

        Parameters
        -----------
        radios: `pycuda.gpuarray.GPUArray`
            Radios chunk
            If not provided, a new GPU array is created.
        clip_min: float, optional
            Before taking the logarithm, the values are clipped to this minimum.
        clip_max: float, optional
            Before taking the logarithm, the values are clipped to this maximum.
        """
        clip_min = clip_min or self.clip_min
        clip_max = clip_max or self.clip_max
        if radios.flags.c_contiguous:
            self.nlog_kernel(
                radios,
                self._nx,
                self._ny,
                self._nz,
                clip_min,
                clip_max,
                grid=self._nblocks,
                block=self._nthreadsperblock
            )
        else:
            # map-like operations cannot be directly applied on 3D arrays
            # that are not C-contiguous. We have to process image per image.
            nz = np.int32(1)
            nthreadsperblock = (32, 32, 1)
            nblocks = tuple([
                updiv(n, p)
                for n, p in zip(
                    [int(self._nx), int(self._ny), int(nz)], nthreadsperblock
                )
            ])
            for i in range(radios.shape[0]):
                self.nlog_kernel(
                    radios[i],
                    self._nx,
                    self._ny,
                    nz,
                    clip_min,
                    clip_max,
                    grid=nblocks,
                    block=nthreadsperblock
                )
        return radios




import os
import numpy as np
from silx.io.url import DataUrl
from tomoscan.esrf.edfscan import EDFTomoScan
from tomoscan.esrf.hdf5scan import HDF5TomoScan
from ..thirdparty.tomwer_load_flats_darks import get_flats_frm_process_file, get_darks_frm_process_file
from .nxflatfield import NXFlatField
from .utils import is_hdf5_extension

dataset_infos = {
    "num_radios": None,
    "num_darks": None,
    "num_flats": None,
    "radios": None,
    "darks": None,
    "flats": None,
    "frame_dims": None,
    "energy_kev": None,
    "distance_m": None,
    "pixel_size_microns": None,
}


class DatasetAnalyzer(object):
    """
    Base class for datasets analyzers.
    """
    def __init__(self, location, processes_file=None, extra_options=None):
        """
        Initialize a Dataset analyzer.

        Parameters
        ----------
        location: str
            Dataset location (directory or file name)
        processes_file: str, optional
            Processes file providing supplementary information (ex. pre-processed data)
        extra_options: dict, optional
            Extra options on how to interpret the dataset.
            Available options are the following:
              - force_flatfield
        """
        self.location = location
        self.processes_file = processes_file
        self._set_extra_options(extra_options)


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "force_flatfield": False,
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options


    def _init_dataset_scan(self, filetype, **kwargs):
        scanners = {
            "edf": EDFTomoScan,
            "hdf5": HDF5TomoScan,
        }
        if filetype not in scanners.keys():
            raise ValueError("No scanner for file type: %s" % filetype)
        scanner = scanners[filetype]
        self.dataset_scanner = scanner(self.location, **kwargs)
        self.projections = self.dataset_scanner.projections
        self.flats = self.dataset_scanner.flats
        self.darks = self.dataset_scanner.darks
        self.n_angles = self.dataset_scanner.tomo_n
        self.radio_dims = (self.dataset_scanner.dim_1, self.dataset_scanner.dim_2)
        self._binning = (1, 1)
        self.translations = None
        self.axis_position = None
        self._radio_dims_notbinned = self.radio_dims

    @property
    def energy(self):
        """
        Return the energy in kev.
        """
        return self.dataset_scanner.energy

    @property
    def distance(self):
        """
        Return the sample-detector distance in meters.
        """
        return abs(self.dataset_scanner.distance)

    @property
    def pixel_size(self):
        """
        Return the pixel size in microns.
        """
        raise ValueError("Must be implemented by inheriting class")

    @property
    def binning(self):
        """
        Return the binning in (x, y)
        """
        return self._binning

    @property
    def rotation_angles(self):
        """
        Return the rotation angles in radians.
        """
        return self._get_rotation_angles()


    def remove_unused_radios(self):
        """
        Remove "unused" radios.
        This is used for legacy ESRF scans.
        """
        # This should only be used with EDF datasets
        if self.dataset_scanner.type != "edf":
            return
        # Extraneous projections are assumed to be on the end
        projs_indices = sorted(self.projections.keys())
        used_radios_range = range(projs_indices[0], self.n_angles)
        radios_not_used = []
        for idx in self.projections.keys():
            if idx not in used_radios_range:
                radios_not_used.append(idx)
        for idx in radios_not_used:
            self.projections.pop(idx)
        return radios_not_used


class EDFDatasetAnalyzer(DatasetAnalyzer):
    """
    EDF Dataset analyzer for legacy ESRF acquisitions
    """
    def __init__(self, location, n_frames=1, processes_file=None, extra_options=None):
        """
        EDF Dataset analyzer.

        Parameters
        -----------
        location: str
            Location of the folder containing EDF files
        """
        super().__init__(location, processes_file=processes_file, extra_options=extra_options)
        if not(os.path.isdir(location)):
            raise ValueError("%s is not a directory" % location)
        self._init_dataset_scan("edf", n_frames=n_frames)

    def _get_rotation_angles(self):
        # Not available in EDF dataset
        return None

    @property
    def pixel_size(self):
        """
        Return the pixel size in microns.
        """
        # TODO X and Y pixel size
        return self.dataset_scanner.pixel_size * 1e6



class HDF5DatasetAnalyzer(DatasetAnalyzer):
    """
    HDF5 dataset analyzer
    """
    def __init__(self, location, processes_file=None, extra_options=None):
        """
        HDF5 Dataset analyzer.

        Parameters
        -----------
        location: str
            Location of the HDF5 master file
        """
        super().__init__(location, processes_file=processes_file, extra_options=extra_options)
        if not(os.path.isfile(location)):
            raise ValueError("%s is not a file" % location)
        self._init_dataset_scan("hdf5")
        self._get_flats_darks()
        self._rot_angles = None


    def _get_flats_darks(self):
        if len(self.flats) == 0 and not(self.extra_options["force_flatfield"]):
            # No flats at all in the dataset. Do nothing.
            return
        if self.processes_file is None and self._load_flats_from_tomwer():
            # Loaded from tomwer_processes.h5
            return
        # Otherwise load or compute flats/darks with nabu
        self._compute_or_load_flats()


    def _load_flats_from_tomwer(self, tomwer_processes_fname=None):
        tomwer_processes_fname = tomwer_processes_fname or "tomwer_processes.h5"
        tomwer_processes_file = os.path.join(self.dataset_scanner.path, "tomwer_processes.h5")
        if not(os.path.isfile(tomwer_processes_file)):
            return False
        print("Loading darks and refs from %s" % tomwer_processes_file)
        new_flats = get_flats_frm_process_file(
            tomwer_processes_file, self.dataset_scanner.entry
        )
        new_darks = get_darks_frm_process_file(
            tomwer_processes_file, self.dataset_scanner.entry
        )
        self.flats = new_flats
        self.darks = new_darks
        return True


    def _compute_or_load_flats(self):
        processes_file = self.processes_file
        if processes_file is None:
            processes_file = os.path.join(self.dataset_scanner.path, "nabu_processes.h5")
        # By default, write in processes_file. Not sure it is a good idea
        results_file = processes_file
        # Don't write in processes_file if flatfield = forced
        if self.extra_options["force_flatfield"]:
            results_file = None
        lookup_files = [
            DataUrl(
                file_path=processes_file,
                data_path=os.path.join(self.dataset_scanner.entry, "flat_field_images")
            )
        ]
        self._nxflatfield = NXFlatField(
            self._get_dataset_hdf5_url(),
            self.dataset_scanner.image_key,
            lookup_files=lookup_files,
            results_file=results_file,
            force_load_existing_results=self.extra_options["force_flatfield"],
            flats_reduction="median",
            darks_reduction="mean"
        )
        res = self._nxflatfield.get_final_urls()
        self.flats = res["flats"]
        self.darks = res["darks"]


    def _get_rotation_angles(self):
        if self._rot_angles is None:
            angles = np.array(self.dataset_scanner.rotation_angle)
            projs_idx = np.array(list(self.projections.keys()))
            angles = angles[projs_idx]
            self._rot_angles = np.deg2rad(angles)
        return self._rot_angles


    def _get_dataset_hdf5_url(self):
        first_proj_idx = sorted(self.projections.keys())[0]
        first_proj_url = self.projections[first_proj_idx]
        return DataUrl(
            file_path=first_proj_url.file_path(),
            data_path=first_proj_url.data_path(),
            data_slice=None
        )


    @property
    def dataset_hdf5_url(self):
        return self._get_dataset_hdf5_url()


    @property
    def pixel_size(self):
        """
        Return the pixel size in microns.
        """
        # TODO X and Y pixel size
        try:
            xs, yz = self.dataset_scanner._get_x_y_magnified_pixel_values()
            ps = xs * 1e6
        except AttributeError:
            ps = self.dataset_scanner.pixel_size * 1e6
        return ps




def analyze_dataset(dataset_path, processes_file=None, extra_options=None):
    if not(os.path.isdir(dataset_path)):
        if not(os.path.isfile(dataset_path)):
            raise ValueError("Error: %s no such file or directory" % dataset_path)
        if not(is_hdf5_extension(os.path.splitext(dataset_path)[-1].replace(".", ""))):
            raise ValueError("Error: expected a HDF5 file")
        dataset_analyzer_class = HDF5DatasetAnalyzer
    else: # directory -> assuming EDF
        dataset_analyzer_class = EDFDatasetAnalyzer
    dataset_structure = dataset_analyzer_class(
        dataset_path, processes_file=processes_file, extra_options=extra_options
    )
    return dataset_structure




import os
import warnings
from time import time
from itertools import product
import numpy as np
from silx.utils.enum import Enum as _Enum


def nextpow2(N, dtype=np.int):
    return 2 ** np.ceil(np.log2(N)).astype(dtype)


def previouspow2(N, dtype=np.int):
    return 2 ** np.floor(np.log2(N)).astype(dtype)


def updiv(a, b):
    return (a + (b - 1)) // b


def convert_index(idx, idx_max, default_val):
    """
    Convert an index (possibly negative or None) to a non-negative integer.

    Parameters
    ----------
    idx: int or None
        Index
    idx_max: int
        Maximum value (upper bound) for the index.
    default_val: int
        Default value if idx is None

    Examples
    ---------
    Given an integer `M`, `J = convert_index(i, M, XX)` returns an integer in the
    mathematical range [0, M] (or Python `range(0, M)`). `J` can then be used
    to define an upper bound of a range.
    """
    if idx is None:
        return default_val
    if idx > idx_max:
        return idx_max
    if idx < 0:
        return idx % idx_max
    return idx


def get_folder_path(foldername=""):
    _file_dir = os.path.dirname(os.path.realpath(__file__))
    package_dir = _file_dir
    return os.path.join(package_dir, foldername)


def get_cuda_srcfile(filename):
    src_relpath = os.path.join("cuda", "src")
    cuda_src_folder = get_folder_path(foldername=src_relpath)
    return os.path.join(cuda_src_folder, filename)


def get_resource_file(filename, subfolder=None):
    subfolder = subfolder or []
    relpath = os.path.join("resources", *subfolder)
    abspath = get_folder_path(foldername=relpath)
    return os.path.join(abspath, filename)


def _sizeof(Type):
    """
    return the size (in bytes) of a scalar type, like the C behavior
    """
    return np.dtype(Type).itemsize


def get_ftype(url):
    """
    return supposed filetype of an url
    """
    if hasattr(url, "file_path"):
        return os.path.splitext(url.file_path())[-1].replace(".", "")
    else:
        return os.path.splitext(url)[-1].replace(".", "")


def get_2D_3D_shape(shape):
    if len(shape) == 2:
        return (1, ) + shape
    return shape


def is_device_backend(backend):
    return backend.lower() in ["cuda", "opencl"]


def get_decay(curve, cutoff=1e3, vmax=None):
    """
    Assuming a decreasing curve, get the first point below a certain threshold.

    Parameters
    ----------
    curve: numpy.ndarray
        A 1D array
    cutoff: float, optional
        Threshold. Default is 1000.
    vmax: float, optional
        Curve maximum value
    """
    if vmax is None:
        vmax = curve.max()
    return np.argmax(np.abs(curve) < vmax / cutoff)


def generate_powers():
    """
    Generate a list of powers of [2, 3, 5, 7],
    up to (2**15)*(3**9)*(5**6)*(7**5).
    """
    primes = [2, 3, 5, 7]
    maxpow = {2: 15, 3: 9, 5: 6, 7: 5}
    valuations = []
    for prime in primes:
        # disallow any odd number (for R2C transform), and any number
        # not multiple of 4 (Ram-Lak filter behaves strangely when
        # dwidth_padded/2 is not even)
        minval = 2 if prime == 2 else 0
        valuations.append(range(minval, maxpow[prime]+1))
    powers = product(*valuations)
    res = []
    for pw in powers:
        res.append(np.prod(list(map(lambda x: x[0] ** x[1], zip(primes, pw)))))
    return np.unique(res)


def calc_padding_lengths1D(length, length_padded):
    """
    Compute the padding lengths at both side along one dimension.

    Parameters
    ----------
    length: int
        Number of elements along one dimension of the original array
    length_padded: tuple
        Number of elements along one dimension of the padded array

    Returns
    -------
    pad_lengths: tuple
        A tuple under the form (padding_left, padding_right). These are the
        lengths needed to pad the original array.
    """
    pad_left = (length_padded - length)//2
    pad_right = length_padded - length - pad_left
    return (pad_left, pad_right)


def calc_padding_lengths(shape, shape_padded):
    """
    Multi-dimensional version of calc_padding_lengths1D.
    Please refer to the documentation of calc_padding_lengths1D.
    """
    assert len(shape) == len(shape_padded)
    padding_lengths = []
    for dim_len, dim_len_padded in zip(shape, shape_padded):
        pad0, pad1 = calc_padding_lengths1D(dim_len, dim_len_padded)
        padding_lengths.append((pad0, pad1))
    return tuple(padding_lengths)


# TODO merge with resources.params
class PaddingMode(_Enum):
    ZEROS = 'zeros'
    MEAN = 'mean'
    EDGE = 'edge'
    SYMMETRIC = 'symmetric'
    REFLECT = 'reflect'


def partition_dict(dict_, n_partitions):
    keys = np.sort(list(dict_.keys()))
    res = []
    for keys_arr in np.array_split(keys, n_partitions):
        d = {}
        for key in keys_arr:
            d[key] = dict_[key]
        res.append(d)
    return res


def subsample_dict(dic, subsampling_factor):
    """
    Subsample a dict where keys are integers.
    """
    res = {}
    indices = sorted(dic.keys())
    for i in indices[::subsampling_factor]:
        res[i] = dic[i]
    return res


def check_supported(param_value, available, param_desc):
    if param_value not in available:
        raise ValueError("Unsupported %s '%s'. Available are: %s" % (param_desc, param_value, str(available)))


def check_supported_enum(param_value, enum_cls, param_desc):
    available = enum_cls.values()
    return check_supported(param_value, available, param_desc)


def check_shape(shape, expected_shape, name):
    if shape != expected_shape:
        raise ValueError(
            "Expected %s shape %s but got %s"
            % (name, str(expected_shape), str(shape))
        )


def copy_dict_items(dict_, keys):
    """
    Perform a shallow copy of a subset of a dictionary.
    The subset if done by a list of keys.
    """
    res = {
        key: dict_[key]
        for key in keys
    }
    return res


class PlaceHolder:
    """
    Base class for place holders.
    """

    def __init__(self):
        pass


class DataPlaceHolder(PlaceHolder):
    """
    Placeholder for data-like objects
    """

    def __init__(self):
        super().__init__()


class ArrayPlaceHolder(PlaceHolder):
    """
    Placeholder for array-like objects
    """

    def __init__(self, shape, dtype, name=None):
        self.shape = shape
        self.dtype = np.dtype(dtype)
        self.data = None
        self.name = name

    @property
    def size(self):
        return np.prod(self.shape)

    @property
    def nbytes(self):
        return self.size * self.dtype.itemsize

    def __repr__(self):
        return "ArrayPlaceHolder(shape=%s, dtype=%s)" % (self.shape, self.dtype)


def array_tostring(arr, show_dtype=False):
    """
    Displays an array without displaying its content.
    """
    if isinstance(arr, ArrayPlaceHolder):
        return arr.__repr__()
    attrs = "shape=%s" % (str(arr.shape))
    if show_dtype:
        attrs += ", dtype=%s" % (str(arr.dtype))
    if hasattr(arr, "gpudata"):  # pycuda.gpuarray
        name = "CudaArray"
        attrs += ", data=%s" % str(arr.gpudata)
        if arr.base is not None:
            attrs += ", base=%s" % arr.base.gpudata
    elif hasattr(arr, "queue"):  # pyopencl.array
        name = "OpenclArray"
        attrs += ", data=%s" % str(arr.data)
    else:  # assuming numpy
        name = "NumpyArray"
        attrs += ", data=%s" % hex(arr.ctypes.data)
    return "%s(%s)" % (name, attrs)


# ------------------------------------------------------------------------------
# ------------------------ Image (move elsewhere ?) ----------------------------
# ------------------------------------------------------------------------------


def generate_coords(img_shp, center=None):
    l_r, l_c = float(img_shp[0]), float(img_shp[1])
    R, C = np.mgrid[:l_r, :l_c]  # np.indices is faster
    if center is None:
        center0, center1 = l_r / 2., l_c / 2.
    else:
        center0, center1 = center
    R += 0.5 - center0
    C += 0.5 - center1
    return R, C


def clip_circle(img, center=None, radius=None):
    R, C = generate_coords(img.shape, center)
    M = R**2+C**2
    res = np.zeros_like(img)
    res[M < radius ** 2] = img[M < radius ** 2]
    return res


def extend_image_onepixel(img):
    # extend of one pixel
    img2 = np.zeros((img.shape[0]+2, img.shape[1]+2), dtype=img.dtype)
    img2[0, 1:-1] = img[0]
    img2[-1, 1:-1] = img[-1]
    img2[1:-1, 0] = img[:, 0]
    img2[1:-1, -1] = img[:, -1]
    # middle
    img2[1: -1, 1:-1] = img
    # corners
    img2[0, 0] = img[0, 0]
    img2[-1, 0] = img[-1, 0]
    img2[0, -1] = img[0, -1]
    img2[-1, -1] = img[-1, -1]
    return img2


def median2(img):
    """
    3x3 median filter for 2D arrays, with "reflect" boundary mode.
    Roughly same speed as scipy median filter, but more memory demanding.
    """
    img2 = extend_image_onepixel(img)
    I = np.array([
        img2[0:-2, 0:-2],
        img2[0:-2, 1:-1],
        img2[0:-2, 2:],
        img2[1:-1, 0:-2],
        img2[1:-1, 1:-1],
        img2[1:-1, 2:],
        img2[2:, 0:-2],
        img2[2:, 1:-1],
        img2[2:, 2:]
    ])
    return np.median(I, axis=0)


# ------------------------------------------------------------------------------
# ---------------------------- Decorators --------------------------------------
# ------------------------------------------------------------------------------

_warnings = {}


def measure_time(func):
    def wrapper(*args, **kwargs):
        t0 = time()
        res = func(*args, **kwargs)
        el = time() - t0
        return el, res
    return wrapper


def wip(func):
    def wrapper(*args, **kwargs):
        func_name = func.__name__
        if func_name not in _warnings:
            _warnings[func_name] = 1
            print("Warning: function %s is a work in progress, it is likely to change in the future")
        return func(*args, **kwargs)
    return wrapper


def warning(msg):
    def decorator(func):
        def wrapper(*args, **kwargs):
            func_name = func.__name__
            if func_name not in _warnings:
                _warnings[func_name] = 1
                print(msg)
                res = func(*args, **kwargs)
                return res
        return wrapper
    return decorator



def deprecated(msg):
    def decorator(func):
        deprecation_warning(msg)
        def wrapper(*args, **kwargs):
            res = func(*args, **kwargs)
            return res
        return wrapper
    return decorator



def deprecation_warning(message):
    warnings.warn(message, DeprecationWarning)





from .validators import *
from .. import __version__

#
# option "type":
#  - required: always visible, user must provide a valid value
#  - optional: visible, but might be left blank
#  - advanced: optional and not visible by default
#  - unsupported: hidden (not implemented yet)
_options_levels = {
    "required": 0,
    "optional": 1,
    "advanced": 2,
    "unsupported": 10,
}


nabu_config = {
    "dataset": {
        "location": {
            "default": "",
            "help": "Dataset location, either a directory or a HDF5-Nexus file.",
            "validator": dataset_location_validator,
            "type": "required",
        },
        "binning": {
            "default": "1",
            "help": "Binning factor in the horizontal dimension when reading the data.\nThe final slices dimensions will be divided by 'slices_binning'",
            "validator": binning_validator,
            "type": "advanced",
        },
        "binning_z": {
            "default": "1",
            "help": "Binning factor in the vertical dimension when reading the data.\nThis retsults in a lesser number of reconstructed slices.",
            "validator": binning_validator,
            "type": "advanced",
        },
        "projections_subsampling": {
            "default": "1",
            "help": "Projections subsampling factor: take one projection out of 'projection_subsampling'",
            "validator": binning_validator,
            "type": "advanced",
        },
    },
    "preproc": {
        "flatfield_enabled": {
            "default": "1",
            "help": "Whether to enable flat-field normalization. If the value is 'forced', then nabu will perform flatfield regardless of the dataset by attempting to load flats/refs from tomwer_processes.h5 file or nabu_processes.h5 or file provided in 'processes_file'.",
            "validator": flatfield_enabled_validator,
            "type": "required",
        },
        "ccd_filter_enabled": {
            "default": "0",
            "help": "Whether to enable the CCD hotspots correction.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "ccd_filter_threshold": {
            "default": "0.04",
            "help": "If ccd_filter_enabled = 1, a median filter is applied on the 3X3 neighborhood\nof every pixel. If a pixel value exceeds the median value more than this parameter,\nthen the pixel value is replaced with the median value.",
            "validator": float_validator,
            "type": "optional",
        },
        "double_flatfield_enabled": {
            "default": "0",
            "help": "Whether to enable the 'double flat-field' filetering for correcting rings artefacts.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "dff_sigma": {
            "default": "",
            "help": "Enable high-pass filtering on double flatfield with this value of 'sigma'",
            "validator": optional_float_validator,
            "type": "advanced",
        },
        "take_logarithm": {
            "default": "1",
            "help": "Whether to take logarithm after flat-field and phase retrieval.",
            "validator": boolean_validator,
            "type": "required",
        },
        "log_min_clip": {
            "default": "1e-6",
            "help": "After division by the FF, and before the logarithm, the is clipped to this minimum. Enabled only if take_logarithm=1",
            "validator": float_validator,
            "type": "advanced",
        },
        "log_max_clip": {
            "default": "10.0",
            "help": "After division by the FF, and before the logarithm, the is clipped to this maximum. Enabled only if take_logarithm=1",
            "validator": float_validator,
            "type": "advanced",
        },
        "sino_normalization": {
            "default": "",
            "help": "Sinogram normalization method. Available methods are: chebyshev, none. Default is none (no normalization)",
            "validator": sino_normalization_validator,
            "type": "advanced",
        },
        "processes_file": {
            "default": "",
            "help": "Load results from a previous session. This the path to a HDF5 file containing results data along with configuration needed to obtain it.",
            "validator": optional_file_location_validator,
            "type": "advanced",
        },
    },
    "phase": {
        "method": {
            "default": "none",
            "help": "Phase retrieval method. Available are: Paganin, None",
            "validator": phase_method_validator,
            "type": "required",
        },
        "delta_beta": {
            "default": "100.0",
            "help": "Single-distance phase retrieval related parameters\n----------------------------\ndelta/beta ratio for the Paganin/CTR method",
            "validator": float_validator,
            "type": "required",
        },
        "margin": {
            "default": "50",
            "help": "Margin (in pixels) in the Paganin/CTR filtering.",
            "validator": integer_validator,
            "type": "advanced",
        },
        "unsharp_coeff": {
            "default": "0",
            "help": "Unsharp mask strength. The unsharped image is equal to\n  UnsharpedImage =  (1 + coeff)*originalPaganinImage - coeff * ConvolvedImage. Setting this coefficient to zero means that no unsharp mask will be applied.",
            "validator": float_validator,
            "type": "optional",
        },
        "unsharp_sigma": {
            "default": "0",
            "help": "Standard deviation of the Gaussian filter when applying an unsharp mask\nafter the Paganin filtering. Disabled if set to 0.",
            "validator": float_validator,
            "type": "optional",
        },
        "padding_type": {
            "default": "edge",
            "help": "Padding type for the filtering step in Paganin/CTR. Available are: mirror, edge, zeros",
            "validator": padding_mode_validator,
            "type": "advanced",
        },
    },
    "reconstruction": {
        "method": {
            "default": "FBP",
            "help": "Reconstruction method. Possible values: FBP, none. If value is 'none', no reconstruction will be done.",
            "validator": reconstruction_method_validator,
            "type": "required",
        },
        "angles_file": {
            "default": "",
            "help": "In the case you want to override the angles found in the files metadata. The angles are in degree.",
            "validator": optional_file_location_validator,
            "type": "optional",
        },
        "rotation_axis_position": {
            "default": "",
            "help": "Rotation axis position. Default is the middle of the detector width. If set to 'auto', nabu will attempt to determine it automatically.",
            "validator": cor_validator,
            "type": "required",
        },
        "axis_correction_file": {
            "default": "",
            "help": "In the case where the axis position is specified for each slice",
            "validator": optional_values_file_validator,
            "type": "advanced",
        },
        "translation_movements_file": {
            "default": "",
            "help": "A file where each line describes the translation in X and Z of the sample (or detector).",
            "validator": optional_values_file_validator,
            "type": "advanced",
        },
        "angle_offset": {
            "default": "0",
            "help": "Use this if you want to obtain a rotated reconstructed slice. The angle is in degrees.",
            "validator": float_validator,
            "type": "advanced",
        },
        "fbp_filter_type": {
            "default": "ramlak",
            "help": "Filter type for FBP method. Available are: ramlak, none",
            "validator": fbp_filter_name_validator,
            "type": "advanced",
        },
        "padding_type": {
            "default": "zeros",
            "help": "Padding type for FBP. Available are: zeros, edges",
            "validator": padding_mode_validator,
            "type": "optional", # put "advanced" with default value "edges" ?
        },
        "enable_halftomo": {
            "default": "0",
            "help": "Whether to enable half-acquisition",
            "validator": boolean_validator,
            "type": "optional",
        },
        "start_x": {
            "default": "0",
            "help": "\nParameters for sub-volume reconstruction. Indices start at 0 !\n----------------------------------------------------------------\n(x, y) are the dimension of a slice, and (z) is the 'vertical' axis\nBy default, all the volume is reconstructed slice by slice, along the axis 'z'.",
            "validator": nonnegative_integer_validator,
            "type": "optional",
        },
        "end_x": {
            "default": "-1",
            "help": "",
            "validator": integer_validator,
            "type": "optional",
        },
        "start_y": {
            "default": "0",
            "help": "",
            "validator": nonnegative_integer_validator,
            "type": "optional",
        },
        "end_y": {
            "default": "-1",
            "help": "",
            "validator": integer_validator,
            "type": "optional",
        },
        "start_z": {
            "default": "0",
            "help": "",
            "validator": nonnegative_integer_validator,
            "type": "optional",
        },
        "end_z": {
            "default": "-1",
            "help": "",
            "validator": integer_validator,
            "type": "optional",
        },
        "iterations": {
            "default": "200",
            "help": "\nParameters for iterative algorithms\n------------------------------------\nNumber of iterations",
            "validator": nonnegative_integer_validator,
            "type": "unsupported",
        },
        "optim_algorithm": {
            "default": "chambolle-pock",
            "help": "Optimization algorithm for iterative methods",
            "validator": optimization_algorithm_name_validator,
            "type": "unsupported",
        },
        "weight_tv": {
            "default": "1.0e-2",
            "help": "Total Variation regularization parameter for iterative methods",
            "validator": float_validator,
            "type": "unsupported",
        },
        "preconditioning_filter": {
            "default": "1",
            "help": "Whether to enable 'filter preconditioning' for iterative methods",
            "validator": boolean_validator,
            "type": "unsupported",
        },
        "positivity_constraint": {
            "default": "1",
            "help": "Whether to enforce a positivity constraint in the reconstruction.",
            "validator": boolean_validator,
            "type": "unsupported",
        },
    },
    "output": {
        "location": {
            "default": "",
            "help": "Directory where the output reconstruction is stored.",
            "validator": optional_directory_location_validator,
            "type": "required",
        },
        "file_prefix": {
            "default": "",
            "help": "File prefix. Optional, by default it is inferred from the scanned dataset.",
            "validator": optional_file_name_validator,
            "type": "optional",
        },
        "file_format": {
            "default": "hdf5",
            "help": "Output file format. Available are: hdf5, tiff, jp2",
            "validator": output_file_format_validator,
            "type": "optional",
        },
        "overwrite_results": {
            "default": "0",
            "help": "What to do in the case where the output file exists.\nBy default, the output data is never overwritten and the process is interrupted if the file already exists.\nSet this option to 1 if you want to overwrite the output files.",
            "validator": boolean_validator,
            "type": "required",
        },
    },
    "postproc": {
        "output_histogram": {
            "default": "0",
            "help": "Whether to compute a histogram of the volume.",
            "validator": boolean_validator,
            "type": "optional",
        },
        "histogram_bins": {
            "default": "1000000",
            "help": "Number of bins for the output histogram. Default is one million. ",
            "validator": nonnegative_integer_validator,
            "type": "advanced",
        },
    },
    "resources": {
        "method": {
            "default": "local",
            "help": "Computations distribution method. It can be:\n  - local:  run the computations on the local machine\n  - slurm: run the computations through SLURM\n  - preview: reconstruct the slices/volume as quickly as possible, possibly doing some binning.",
            "validator": distribution_method_validator,
            "type": "required",
        },
        "gpus": {
            "default": "1",
            "help": "Number of GPUs to use.",
            "validator": nonnegative_integer_validator,
            "type": "unsupported",
        },
        "gpu_id": {
            "default": "",
            "help": "For method = local only. List of GPU IDs to use. This parameter overwrites 'gpus'.\nIf left blank, exactly one GPU will be used, and the best one will be picked.",
            "validator": list_of_int_validator,
            "type": "unsupported",
        },
        "cpu_workers": {
            "default": "0",
            "help": "Number of 'CPU workers' for each GPU worker. It is discouraged to set this number to more than one. A value of -1 means exactly one CPU worker.",
            "validator": integer_validator,
            "type": "unsupported",
        },
        "memory_per_node": {
            "default": "90%",
            "help": "RAM memory per computing node, either in GB or in percent of the AVAILABLE (!= total) node memory.\nIf several workers share the same node, their combined memory usage will not exceed this number.",
            "validator": resources_validator,
            "type": "unsupported",
        },
        "threads_per_node": {
            "default": "100%",
            "help": "Number of threads to allocate on each node, either a number or a percentage of the available threads",
            "validator": resources_validator,
            "type": "unsupported",
        },
        "queue": {
            "default": "gpu",
            "help": "\nParameters exclusive to the 'slurm' distribution method\n------------------------------------------------------\nName of the SLURM partition ('queue'). Full list is obtained with 'scontrol show partition'",
            "validator": nonempty_string_validator,
            "type": "unsupported",
        },
        "walltime": {
            "default": "01:00:00",
            "help": "Time limit for the SLURM resource allocation, in the format Hours:Minutes:Seconds",
            "validator": walltime_validator,
            "type": "unsupported",
        },
    },
    "about": {
        "nabu_version": {
            "default": __version__,
            "help": "Version of the nabu software",
            "validator": no_validator,
            "type": "required",
        },
        "verbosity": {
            "default": "2",
            "help": "Level of verbosity of the processing. 0 = terse, 3 = much information.",
            "validator": logging_validator,
            "type": "optional",
        },
    },
}

renamed_keys = {
    "marge": {
        "section": "phase",
        "new_name": "margin",
        "since": "2020.2.0",
        "message": "Option 'marge' has been renamed to 'margin' in [phase]",
    },
    "overwrite_results": {
        "section": "about",
        "new_name": "overwrite_results",
        "new_section": "output",
        "since": "2020.3.0",
        "message": "Option 'overwrite_results' was moved from section [about] to section [output]",
    },
    "nabu_config_version": {
        "section": "about",
        "new_name": "",
        "new_section": "about",
        "since": "2020.3.1",
        "message": "Option 'nabu_config_version' was removed.",
    },
}



import os
from ..utils import PlaceHolder, DataPlaceHolder, copy_dict_items
from ..io.config import NabuConfigParser, validate_nabu_config
from .utils import is_hdf5_extension
from .dataset_analyzer import analyze_dataset, EDFDatasetAnalyzer, HDF5DatasetAnalyzer
from .dataset_validator import NabuValidator
from .cor import CORFinder


class ProcessConfig:
    """
    A class for describing the Nabu process configuration.
    """

    def __init__(
        self,
        conf_fname=None,
        conf_dict=None,
        dataset_infos=None,
        checks=True,
        remove_unused_radios=True,
    ):
        """
        Initialize a ProcessConfig class.

        Parameters
        ----------
        conf_fname: str
            Path to the nabu configuration file. If provided, the parameters
            `conf_dict` and `dataset_infos` are ignored.
        conf_dict: dict
            A dictionary describing the nabu processing steps.
            If provided, it should be provided along with `dataset_infos` ; in this
            case, the parameter `conf_fname` is ignored.
        dataset_infos: EDFDatasetAnalyzer
            A `EDFDatasetAnalyzer` class instance.
            If provided, it should be provided along with `conf_dict` ; in this
            case, the parameter `conf_fname` is ignored.
        checks: bool, optional, default is True
            Whether to perform checks on configuration and datasets (recommended !)
        remove_unused_radios: bool, optional, default is True
            Whether to remove unused radios, i.e radios present in the dataset,
            but not explicitly listed in the scan metadata.
        """
        args_error_msg = (
            "You must either provide conf_fname, or (conf_dict and dataset_infos)"
        )
        self.conf_fname = conf_fname
        if conf_fname is not None:
            if (conf_dict is not None) or (dataset_infos is not None):
                raise ValueError(args_error_msg)
            if not (os.path.isfile(conf_fname)):
                raise ValueError("No such file: %s" % conf_fname)
            conf = NabuConfigParser(conf_fname).conf_dict
            conf = validate_nabu_config(conf)
            dataset_infos = analyze_dataset(
                conf["dataset"]["location"],
                processes_file=conf["preproc"]["processes_file"],
                extra_options={
                    "force_flatfield": conf["preproc"]["flatfield_enabled"] == "forced"
                }
            )
        else:
            if (conf_dict is None) or (dataset_infos is None):
                raise ValueError(args_error_msg)
            conf = validate_nabu_config(conf_dict)
            assert (isinstance(dataset_infos, EDFDatasetAnalyzer)) or (
                isinstance(dataset_infos, HDF5DatasetAnalyzer)
            )
        self.nabu_config = conf
        self.dataset_infos = dataset_infos
        self.checks = checks
        self.remove_unused_radios = remove_unused_radios
        self._get_cor()
        self.validation_stage2()
        self.build_processing_steps()


    def _get_cor(self):
        cor = self.nabu_config["reconstruction"]["rotation_axis_position"]
        if isinstance(cor, str): # auto-CoR
            self.corfinder = CORFinder(
                self.dataset_infos,
                halftomo=self.nabu_config["reconstruction"]["enable_halftomo"],
                do_flatfield=self.nabu_config["preproc"]["flatfield_enabled"]
            )
            cor = self.corfinder.find_cor(search_method=cor)
        self.dataset_infos.axis_position = cor


    def validation_stage2(self):
        validator = NabuValidator(self.nabu_config, self.dataset_infos)
        if self.checks:
            validator.perform_all_checks(remove_unused_radios=self.remove_unused_radios)


    def build_processing_steps(self):
        """
        Build a list of processing steps from a ProcessConfig instance.
        The returned structures are a more compact and ready-to-use representation
        of the two main fields of ProcessConfig (dataset_infos and nabu_config).
        """
        nabu_config = self.nabu_config
        dataset_infos = self.dataset_infos
        binning = (nabu_config["dataset"]["binning"], nabu_config["dataset"]["binning_z"])
        tasks = []
        options = {}

        #
        # Dataset / Get data
        #
        # First thing to do is to get the data (radios or sinograms)
        # For now data is assumed to be on disk (see issue #66).
        tasks.append("read_chunk")
        options["read_chunk"] = {
            "files": dataset_infos.projections, # TODO handle sinograms
            "sub_region": None,
            "binning": binning,
            "dataset_subsampling": nabu_config["dataset"]["projections_subsampling"]
        }
        #
        # Flat-field
        #
        if nabu_config["preproc"]["flatfield_enabled"]:
            tasks.append("flatfield")
            options["flatfield"] = {
                #  ChunkReader handles binning/subsampling by itself,
                # but FlatField needs "real" indices (after binning/subsampling)
                "projs_indices": dataset_infos._projs_indices_subsampled,
                "binning": binning,
            }
        if nabu_config["preproc"]["ccd_filter_enabled"]:
            tasks.append("ccd_correction")
            options["ccd_correction"] = {
                "type": "median_clip", # only one available for now
                "median_clip_thresh": nabu_config["preproc"]["ccd_filter_threshold"],
            }
        #
        # Double flat field
        #
        if nabu_config["preproc"]["double_flatfield_enabled"]:
            tasks.append("double_flatfield")
            options["double_flatfield"] = {
                # ~ "results_location": nabu_config["output"]["location"], # not useful (?)
                "sigma": nabu_config["preproc"]["dff_sigma"],
            }
        #
        #
        # Phase retrieval
        #
        if nabu_config["phase"]["method"] != None:
            tasks.append("phase")
            options["phase"] = copy_dict_items(
                nabu_config["phase"],
                ["delta_beta", "margin", "padding_type"]
            )
            options["phase"].update({
                "energy_kev": dataset_infos.energy,
                "distance_cm": dataset_infos.distance * 1e2,
                "pixel_size_microns": dataset_infos.pixel_size,
            })
            if binning != (1, 1):
                options["phase"]["delta_beta"] /= (binning[0] * binning[1])
        #
        # Unsharp
        #
        if nabu_config["phase"]["unsharp_coeff"] > 0:
            tasks.append("unsharp_mask")
            options["unsharp_mask"] = copy_dict_items(
                nabu_config["phase"], ["unsharp_coeff", "unsharp_sigma"]
            )
        #
        # -logarithm
        #
        if nabu_config["preproc"]["take_logarithm"]:
            tasks.append("take_log")
            options["take_log"] = copy_dict_items(nabu_config["preproc"], ["log_min_clip", "log_max_clip"])
        #
        # Translation movements
        #
        translations = dataset_infos.translations
        if translations is not None:
            tasks.append("radios_movements")
            options["radios_movements"] = {
                "translation_movements": dataset_infos.translations
            }
        #
        # Sinogram normalization (before half-tomo)
        #
        if nabu_config["preproc"]["sino_normalization"] != None:
            tasks.append("sino_normalization")
            options["sino_normalization"] = {
                "method": nabu_config["preproc"]["sino_normalization"]
            }
        #
        # Reconstruction
        #
        if nabu_config["reconstruction"]["method"] != None:
            tasks.append("build_sino")
            options["build_sino"] = copy_dict_items(
                nabu_config["reconstruction"],
                ["rotation_axis_position", "enable_halftomo", "start_x", "end_x",
                 "start_y", "end_y", "start_z", "end_z"]
            )
            options["build_sino"]["axis_correction"] = dataset_infos.axis_correction
            tasks.append("reconstruction")
            # Iterative is not supported through configuration file for now.
            options["reconstruction"] = copy_dict_items(
                nabu_config["reconstruction"],
                ["method", "rotation_axis_position", "fbp_filter_type",
                "padding_type", "enable_halftomo",
                "start_x", "end_x", "start_y", "end_y", "start_z", "end_z"]
            )
            rec_options = options["reconstruction"]
            rec_options["rotation_axis_position"] = dataset_infos.axis_position
            options["build_sino"]["rotation_axis_position"] = dataset_infos.axis_position
            rec_options["axis_correction"] = dataset_infos.axis_correction
            rec_options["angles"] = dataset_infos.reconstruction_angles
            rec_options["radio_dims_y_x"] = dataset_infos.radio_dims[::-1]
            rec_options["pixel_size_cm"] = dataset_infos.pixel_size * 1e-4 # pix size is in microns
            if rec_options["enable_halftomo"]:
                rec_options["angles"] = rec_options["angles"][:rec_options["angles"].size//2]
                cor_i = int(round(rec_options["rotation_axis_position"]))
                # New keys
                rec_options["rotation_axis_position_halftomo"] = (2*cor_i-1)/2.
            # New key
            rec_options["cor_estimated_auto"] = isinstance(nabu_config["reconstruction"]["rotation_axis_position"], str)

        #
        # Histogram
        #
        if nabu_config["postproc"]["output_histogram"]:
            # Current limitation
            if not(is_hdf5_extension(nabu_config["output"]["file_format"])):
                raise ValueError("Histograms are only supported when output is HDF5")
            #
            tasks.append("histogram")
            options["histogram"] = copy_dict_items(
                nabu_config["postproc"], ["histogram_bins"]
            )

        #
        # Save
        #
        if nabu_config["output"]["location"] is not None:
            tasks.append("save")
            options["save"] = copy_dict_items(
                nabu_config["output"], list(nabu_config["output"].keys())
            )
            options["save"]["overwrite"] = nabu_config["output"]["overwrite_results"]

        # "sub_region" is the same for all steps
        # ~ sub_region = options["read_chunk"]["sub_region"]
        # ~ for step in options.keys():
            # ~ options[step]["sub_region"] = sub_region

        self.processing_steps = tasks
        self.processing_options = options



#!/usr/bin/env python

"""
computations.py: determine computational needs, chunking method to be used, etc.
"""

from silx.image.tomography import get_next_power


def estimate_required_memory(process_config, chunk_size=None, radios_and_slices=True):
    """
    Estimate the memory (RAM) needed for a reconstruction.

    Parameters
    -----------
    radios_and_slices: bool
        Whether radios and slices will co-exist in memory (meaning more memory usage)
    """
    dataset = process_config.dataset_infos
    nabu_config = process_config.nabu_config
    processing_steps = process_config.processing_steps
    Nx, Ny = dataset.radio_dims
    Na = dataset.n_angles
    if chunk_size is not None:
        Ny = chunk_size

    total_memory_needed = 0

    # Read data
    # ----------
    binning_z = nabu_config["dataset"]["binning_z"]
    projections_subsampling = nabu_config["dataset"]["projections_subsampling"]
    data_volume_size = Nx * Ny * Na * 4
    data_image_size = Nx * Ny * 4
    total_memory_needed += data_volume_size

    # CCD processing
    # ---------------
    if "flatfield" in processing_steps:
        # Flat-field is done in-place, but still need to load darks/flats
        n_darks = len(dataset.darks)
        n_flats = len(dataset.flats)
        darks_size = n_darks * Nx * Ny * 2  # uint16
        flats_size = n_flats * Nx * Ny * 4  # f32
        total_memory_needed += darks_size + flats_size

    if "ccd_correction" in processing_steps:
        # CCD filter is "batched 2D"
        total_memory_needed += data_image_size

    # Phase retrieval
    # ---------------
    if "phase" in processing_steps:
        # Phase retrieval is done image-wise, so near in-place, but needs to
        # allocate some images, fft plans, and so on
        Nx_p = get_next_power(2 * Nx)
        Ny_p = get_next_power(2 * Ny)
        img_size_real = 2 * 4 * Nx_p * Ny_p
        img_size_cplx = 2 * 8 * ((Nx_p * Ny_p) // 2 + 1)
        total_memory_needed += 2 * img_size_real + 3 * img_size_cplx

    # Reconstruction
    # ---------------
    reconstructed_volume_size = 0
    if "reconstruction" in processing_steps:
        if radios_and_slices:
            total_memory_needed += data_volume_size  # radios + sinos
        rec_config = process_config.processing_options["reconstruction"]
        if "rotation_axis_position_halftomo" in rec_config:
            # Slice has a different shape in half acquisition.
            # Not sure if start_x, ..., end_y  are supported
            rc = rec_config["rotation_axis_position_halftomo"]
            Nx_rec = 2 * rc
            Nx_rec = max(2 * rc, 2 * (Nx - rc))
            Ny_rec = Nx_rec
        else:
            Nx_rec = (rec_config["end_x"] - rec_config["start_x"] + 1)
            Ny_rec = (rec_config["end_y"] - rec_config["start_y"] + 1)
        Nz_rec = (rec_config["end_z"] - rec_config["start_z"] + 1) // binning_z
        if chunk_size:
            Nz_rec = chunk_size // binning_z
        reconstructed_volume_size = Nx_rec * Ny_rec * Nz_rec * 4  # float32
        total_memory_needed += reconstructed_volume_size

    return total_memory_needed

def estimate_chunk_size(available_memory_GB, process_config, chunk_step=50):
    """
    Estimate the maximum chunk size given an avaiable amount of memory.

    Parameters
    -----------
    available_memory_GB: float
        available memory in Giga Bytes (GB - not GiB !).
    process_config: ProcessConfig
        Instance of nabu.resources.processconfig.ProcessConfig
    """
    # Deriving a closed-form formula for estimate_required_memory would be cumbersome
    # so we use this inelegant way instead
    chunk_size = chunk_step
    radios_and_slices = False
    if (
        "reconstruction" in process_config.processing_steps
        and process_config.processing_options["reconstruction"]["enable_halftomo"]
    ):
        radios_and_slices = True

    max_dz = process_config.dataset_infos.radio_dims[1]
    chunk_size = chunk_step
    last_good_chunk_size = chunk_size
    while True:
        required_mem = estimate_required_memory(
            process_config, chunk_size=chunk_size, radios_and_slices=radios_and_slices
        )
        required_mem_GB = required_mem / 1e9
        if required_mem_GB > available_memory_GB or chunk_size > max_dz:
            break
        last_good_chunk_size = chunk_size
        chunk_size += chunk_step
    return last_good_chunk_size



import os.path as path
from os import linesep
from configparser import ConfigParser, MissingSectionHeaderError, DuplicateOptionError
from silx.io.dictdump import dicttoh5, h5todict
from silx.io.url import DataUrl
from ..utils import check_supported
from ..resources.nabu_config import nabu_config, _options_levels, renamed_keys

class NabuConfigParser(object):
    def __init__(self, fname):
        """
        Nabu configuration file parser.

        Parameters
        ----------
        fname: str
            File name of the configuration file
        """
        parser = ConfigParser(
            inline_comment_prefixes=("#",) # allow in-line comments
        )
        with open(fname) as fid:
            file_content = fid.read()
        parser.read_string(file_content)
        self.parser = parser
        self.get_dict()
        self.file_content = file_content.split(linesep)

    def get_dict(self):
        # Is there an officially supported way to do this ?
        self.conf_dict = self.parser._sections

    def __str__(self):
        return self.conf_dict.__str__()

    def __repr__(self):
        return self.conf_dict.__repr__()

    def __getitem__(self, key):
        return self.conf_dict[key]


def generate_nabu_configfile(
    fname,
    config=None, sections=None, comments=True, options_level=None, prefilled_values=None
):
    """
    Generate a nabu configuration file.

    Parameters
    -----------
    fname: str
        Output file path.
    config: dict
        Configuration to save. If section and / or key missing will store the
        default value
    sections: list of str, optional
        Sections which should be included in the configuration file
    comments: bool, optional
        Whether to include comments in the configuration file
    options_level: str, optional
        Which "level" of options to embed in the file. Can be "required", "optional", "advanced".
        Default is "optional".
    """
    if options_level is None:
        options_level = "optional"
    if prefilled_values is None:
        prefilled_values = {}
    check_supported(options_level, list(_options_levels.keys()), "options_level")
    options_level = _options_levels[options_level]
    parser = ConfigParser(
        inline_comment_prefixes=("#",) # allow in-line comments
    )
    if config is None:
        config = {}
    if sections is None:
        sections = nabu_config.keys()
    with open(fname, "w") as fid:
        for section, section_content in nabu_config.items():
            if section not in sections:
                continue
            if section != "dataset":
                fid.write("%s%s" % (linesep, linesep))
            fid.write("[%s]%s" % (section, linesep))
            for key, values in section_content.items():
                if options_level < _options_levels[values["type"]]:
                    continue
                if comments and values["help"].strip() != "":
                    for help_line in values["help"].split(linesep):
                        content = "# %s" % (help_line) if help_line.strip() != "" else ""
                        content = content + linesep
                        fid.write(content)
                value = values["default"]
                if section in prefilled_values and key in prefilled_values[section]:
                    value = prefilled_values[section][key]
                if section in config and key in config[section]:
                    value = config[section][key]
                fid.write("%s = %s%s" % (key, value, linesep))



def _extract_nabuconfig_section(section):
    res = {}
    for key, val in nabu_config[section].items():
        res[key] = val["default"]
    return res


def _extract_nabuconfig_keyvals():
    res = {}
    for section in nabu_config.keys():
        res[section] = _extract_nabuconfig_section(section)
    return res


def _handle_modified_key(key, val, section):
    if val is not None:
        return key, val, section
    if key in renamed_keys and renamed_keys[key]["section"] == section:
        info = renamed_keys[key]
        print(info["message"])
        print(
            "This is deprecated since version %s and will result in an error in futures versions"
            % (info["since"])
        )
        section = info.get("new_section", section)
        if info["new_name"] == "":
            return None, None, section # deleted key
        val = nabu_config[section].get(info["new_name"], None)
        return info["new_name"], val, section
    else:
        return key, None, section # unhandled renamed/deleted key


def validate_nabu_config(config):
    """
    Validate a nabu configuration.

    Parameters
    ----------
    config: str or dict
        Configuration. Can be a dictionary or a path to a configuration file.
    """
    if isinstance(config, str):
        config = NabuConfigParser(config).conf_dict
    res_config = {}
    for section, section_content in config.items():
        # Ignore the "other" section
        if section.lower() == "other":
            continue
        if section not in nabu_config:
            raise ValueError("Unknown section [%s]" % section)
        res_config[section] = _extract_nabuconfig_section(section)
        res_config[section].update(section_content)
        for key, value in res_config[section].items():
            opt = nabu_config[section].get(key, None)
            key, opt, section = _handle_modified_key(key, opt, section)
            if key is None:
                continue # deleted key
            if opt is None:
                raise ValueError("Unknown option '%s' in section [%s]" % (key, section))
            validator = nabu_config[section][key]["validator"]
            res_config[section][key] = validator(section, key, value)
    # Handle sections missing in config
    for section in (set(nabu_config.keys()) - set(res_config.keys())):
        res_config[section] = _extract_nabuconfig_section(section)
        for key, value in res_config[section].items():
            validator = nabu_config[section][key]["validator"]
            res_config[section][key] = validator(section, key, value)
    return res_config



def convert_dict_values(dic, val_replacements, bytes_tostring=False):
    """
    Modify a dictionary to be able to export it with silx.io.dicttoh5
    """
    modified_dic = {}
    for key, value in dic.items():
        if isinstance(key, int): # np.isscalar ?
            key = str(key)
        if isinstance(value, bytes) and bytes_tostring:
            value = bytes.decode(value.tostring())
        if isinstance(value, dict):
            value = convert_dict_values(value, val_replacements, bytes_tostring=bytes_tostring)
        else:
            if isinstance(value, DataUrl):
                value = value.path()
            elif value.__hash__ is not None and value in val_replacements:
                value = val_replacements[value]
        modified_dic[key] = value
    return modified_dic


def export_dict_to_h5(dic, h5file, h5path, overwrite_data=True, mode="a"):
    """
    Wrapper on top of silx.io.dictdump.dicttoh5 replacing None with "None"

    Parameters
    -----------
    dic: dict
        Dictionary containing the options
    h5file: str
        File name
    h5path: str
        Path in the HDF5 file
    overwrite_data: bool, optional
        Whether to overwrite data when writing HDF5. Default is True
    mode: str, optional
        File mode. Default is "a" (append).
    """
    modified_dic = convert_dict_values(
        dic,
        {None: "None"},
    )
    return dicttoh5(
        modified_dic,
        h5file=h5file,
        h5path=h5path,
        overwrite_data=overwrite_data,
        mode=mode
    )


def import_h5_to_dict(h5file, h5path, asarray=False):
    dic = h5todict(h5file, path=h5path, asarray=asarray)
    modified_dic = convert_dict_values(
        dic,
        {"None": None},
        bytes_tostring=True
    )
    return modified_dic




import numpy as np
from .sinogram import SinoProcessing
from ..thirdparty.pore3d_deringer_munch import munchetal_filter
from ..utils import check_supported

class Deringer(SinoProcessing):
    """
    Class aimed at wrapping several sinogram-based rings correction methods.
    """

    _available_methods = {
        "munch": munchetal_filter
    }

    def __init__(self, sinos_shape=None, radios_shape=None, method="munch", deringer_args=None, deringer_kwargs=None):
        """
        Parameters
        -----------
        method: str
            Sinogram rings correction method.
        deringer_args: list or tuple
            List of options to pass (additionally to the current sinogram) to
            the sinogram de-striping function.
        deringer_kwargs: dict
            Dictionary of named options to pass to the sinogram de-striping function.

        Please see the SinoProcessing documentation for other parameters
        """
        super().__init__(sinos_shape=sinos_shape, radios_shape=radios_shape)
        self._init_destriping_function(method, deringer_args, deringer_kwargs)


    def _init_destriping_function(self, method, deringer_args, deringer_kwargs):
        check_supported(method, list(self._available_methods.keys()), "method")
        self.method = method
        self._destriping_func = self._available_methods[method]
        self.deringer_args = []
        if deringer_args is not None:
            self.deringer_args = deringer_args
        self.deringer_kwargs = {}
        if deringer_kwargs is not None:
            self.deringer_kwargs = deringer_kwargs

    def _destripe_sinogram(self, sinogram):
        return self._destriping_func(
            sinogram,
            *self.deringer_args,
            **self.deringer_kwargs
        )


    def correct_rings(self, sinos, output=None):
        """
        Correct the rings in sinogram.
        Defaults to in-place processing !

        Parameters
        -----------
        sino: numpy.ndarray, optional
            Sinogram or stack of sinograms.
        output: numpy.ndarray, optional
            Stack of sinograms. If not provided, the correction will overwrite
            the sinogram passed as input.
        """
        if output is None:
            output = sinos
        # TODO more elegant
        if sinos.ndim == 2:
            output[:] = self._destripe_sinogram(sinos)
            return output
        n_sinos = sinos.shape[0]
        for i in range(n_sinos):
            output[i] = self._destripe_sinogram(sinos[i])
        return output



import os
import numpy as np
from silx.io import get_data
from ..utils import subsample_dict
from .validators import validator, directory_writeable_validator
from .utils import get_values_from_file


"""
The NabuValidator class is a "second-stage validation".
The first stage was:
  - Validate individual entries of the nabu config file / config dict
  - Analyze the dataset and extract information (radios, flats, darks, energy, ...)
Now we couple these two (nabu config and dataset) together in order to cross check.
"""


class NabuValidator(object):
    def __init__(self, nabu_config, dataset_infos):
        """
        Perform a cross-validation of nabu configuration against dataset informations.
        Check the consistency of these two structures, and possibly modify them in-place.

        Parameters
        ----------
        nabu_config: dict
            Dictionary containing the nabu configuration, usually got from
            `nabu.io.config.validate_nabu_config()`
            It will possibly be modified !
        dataset_infos: `DatasetAnalyzer`
            Structure containing information on the dataset to process.
            It will possibly be modified !
        """
        self.nabu_config = nabu_config
        self.dataset_infos = dataset_infos
        self.check_not_empty() # has to be done first
        self.convert_negative_indices()
        self.get_angles()
        self._get_rotation_axis()
        self._get_translation_file()
        self._get_resources()
        self._get_output_filename()


    @staticmethod
    def _convert_negative_idx(idx, last_idx):
        res = idx
        if idx < 0:
            res = last_idx + idx
        return res


    def _get_nx_ny(self):
        if self.nabu_config["reconstruction"]["enable_halftomo"]:
            cor = int(round(self.dataset_infos.axis_position))
            nx = max(2*cor, 2 * (self.dataset_infos.radio_dims[0] - 1 - cor))
        else:
            nx, nz = self.dataset_infos.radio_dims
        ny = nx
        return nx, ny


    def convert_negative_indices(self):
        """
        Convert any negative index to the corresponding positive index.
        """
        nx, nz = self.dataset_infos.radio_dims
        ny = nx
        if self.nabu_config["reconstruction"]["enable_halftomo"]:
            if self.dataset_infos.axis_position is None:
                raise ValueError("Cannot use rotation axis position in the middle of the detector when half tomo is enabled")
            nx, ny = self._get_nx_ny()
        what = (
            ("reconstruction", "start_x", nx),
            ("reconstruction", "end_x", nx),
            ("reconstruction", "start_y", ny),
            ("reconstruction", "end_y", ny),
            ("reconstruction", "start_z", nz),
            ("reconstruction", "end_z", nz),
        )
        for section, key, upper_bound in what:
            self.nabu_config[section][key] = self._convert_negative_idx(
                self.nabu_config[section][key], upper_bound
            )


    def check_not_empty(self):
        assert len(self.dataset_infos.projections) > 0, "Dataset seems to be empty (no projections)"
        assert self.dataset_infos.n_angles is not None, "Could not determine the number of projections. Please check the .info or HDF5 file"
        for dim_name, n in zip(["dim_1", "dim_2"], self.dataset_infos.radio_dims):
            assert n is not None, "Could not determine %s. Please check the .info file or HDF5 file" % dim_name
        self._projs_indices = sorted(self.dataset_infos.projections.keys())


    def get_angles(self):
        rec_params = self.nabu_config["reconstruction"]
        n_angles = self.dataset_infos.n_angles
        angles_file = rec_params["angles_file"]
        if angles_file is not None:
            try:
                angles = get_values_from_file(angles_file, n_values=n_angles)
                angles = np.deg2rad(angles)
            except ValueError:
                print("Something wrong with angle file %s" % angles_file)
                raise
        else:
            angles = self.dataset_infos.rotation_angles
            if angles is None:
                if rec_params["enable_halftomo"]:
                    angles = np.linspace(0, 2*np.pi, n_angles, True)
                else:
                    angles = np.linspace(0, np.pi, n_angles, False)
        angles += np.deg2rad(rec_params["angle_offset"])
        self.dataset_infos.reconstruction_angles = angles


    def _get_rotation_axis(self):
        rec_params = self.nabu_config["reconstruction"]
        axis_correction_file = rec_params["axis_correction_file"]
        axis_correction = None
        if axis_correction_file is not None:
            try:
                axis_correction = get_values_from_file(
                    axis_correction_file, n_values=self.dataset_infos.n_angles
                ).astype(np.flooat32)
            except ValueError:
                print("Something wrong with axis correction file %s" % axis_correction_file)
                raise
        self.dataset_infos.axis_correction = axis_correction


    def _get_translation_file(self):
        rec_params = self.nabu_config["reconstruction"]
        transl_file = rec_params["translation_movements_file"]
        if transl_file in (None, ''):
            return
        translations = None
        if transl_file is not None and "://" not in transl_file:
            try:
                translations = get_values_from_file(
                    transl_file, shape=(self.dataset_infos.n_angles, 2)
                ).astype(np.float32)
            except ValueError:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        else:
            try:
                translations = get_data(transl_file)
            except:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        self.dataset_infos.translations = translations


    def _get_resources(self):
        opts = self.nabu_config["resources"]
        if opts["gpu_id"] != []:
            opts["gpus"] = len(opts["gpu_id"])
        if opts["gpus"] == 0:
            opts["gpu_id"] = []


    def _get_output_filename(self):
        opts = self.nabu_config["output"]
        dataset_path = self.nabu_config["dataset"]["location"]
        if opts["location"] == "" or opts["location"] is None:
            opts["location"] = os.path.dirname(dataset_path)
        if opts["file_prefix"] == "" or opts["file_prefix"] is None:
            if os.path.isfile(dataset_path): # hdf5
                file_prefix = os.path.basename(dataset_path).split(".")[0]
            elif os.path.isdir(dataset_path):
                file_prefix = os.path.basename(dataset_path)
            else:
                raise ValueError(
                    "dataset location %s is neither a file or directory"
                    % dataset_path
                )
            file_prefix += "_rec" # avoid overwriting dataset
            opts["file_prefix"] = file_prefix


    #
    # User-triggered checks
    #
    def check_first_last_radios(self):
        first_idx = self._projs_indices[0]
        last_idx = self._projs_indices[-1]


    def check_can_do_flatfield(self):
        if self.nabu_config["preproc"]["flatfield_enabled"]:
            darks = self.dataset_infos.darks
            assert len(darks) > 0, "Need at least one dark to perform flat-field correction"
            for dark_id, dark_url in darks.items():
                assert os.path.isfile(dark_url.file_path()), "Dark file %s not found" % dark_url.file_path()
            flats = self.dataset_infos.flats
            assert len(flats) > 0, "Need at least one flat to perform flat-field correction"
            for flat_id, flat_url in flats.items():
                assert os.path.isfile(flat_url.file_path()), "Flat file %s not found" % flat_url.file_path()


    @staticmethod
    def _check_start_end_idx(start, end, n_elements, start_name="start_x", end_name="end_x"):
        assert (start >= 0 and start < n_elements), "Invalid value for %s, must be >= 0 and < %d" % (start_name, n_elements)
        assert (end >= 0 and end < n_elements), "Invalid value for %s, must be >= 0 and < %d" % (end_name, n_elements)


    def check_slice_indexes(self):
        nx, nz = self.dataset_infos.radio_dims
        rec_params = self.nabu_config["reconstruction"]
        if rec_params["enable_halftomo"]:
            ny, nx = self._get_nx_ny()
        what = (
            ("start_x", "end_x", nx),
            ("start_y", "end_y", nx),
            ("start_z", "end_z", nz)
        )
        for (start_name, end_name, numels) in what:
            self._check_start_end_idx(
                rec_params[start_name],
                rec_params[end_name],
                numels,
                start_name=start_name,
                end_name=end_name
            )


    def perform_all_checks(self, remove_unused_radios=True):
        """
        Do the final stage of validation: check the nabu configuration against
        the analyzed dataset.
        """
        self.check_first_last_radios()
        self.check_can_do_flatfield()
        self.check_slice_indexes()
        if remove_unused_radios:
            self.remove_unused_radios()
        self.handle_processing_mode()
        self.handle_binning()
        self.check_output_file()


    #
    # Dataset modifications
    #
    def remove_unused_radios(self):
        unused_radios = self.dataset_infos.remove_unused_radios()
        # update
        self._projs_indices = sorted(self.dataset_infos.projections.keys())


    def handle_binning(self):
        """
        Modify the dataset description and nabu config to handle binning and
        projections subsampling.
        """
        self.dataset_infos._radio_dims_notbinned = self.dataset_infos.radio_dims
        dataset_cfg = self.nabu_config["dataset"]
        self.binning = (dataset_cfg["binning"], dataset_cfg["binning_z"])
        self.dataset_infos._binning = self.binning
        subsampling_factor = dataset_cfg["projections_subsampling"]
        self.projections_subsampling = subsampling_factor
        self.dataset_infos._projections_subsampled = self.dataset_infos.projections
        self.dataset_infos._projs_indices_subsampled = self._projs_indices
        if subsampling_factor > 1:
            self.dataset_infos._projections_subsampled = subsample_dict(self.dataset_infos.projections, subsampling_factor)
            self.dataset_infos._projs_indices_subsampled = sorted(self.dataset_infos._projections_subsampled.keys())
            self.dataset_infos.reconstruction_angles = self.dataset_infos.reconstruction_angles[::subsampling_factor]
            self.dataset_infos.n_angles //= subsampling_factor
        if self.binning != (1, 1):
            bin_x, bin_z = self.binning
            nx, nz = self.dataset_infos.radio_dims
            nx //= bin_x
            ny = nx # square slices
            nz //= bin_z
            self.dataset_infos.radio_dims = (nx, nz)
            rec_cfg = self.nabu_config["reconstruction"]
            end_x = rec_cfg["end_x"]
            end_y = rec_cfg["end_y"]
            end_z = rec_cfg["end_z"]
            bin_y = bin_x # square slices
            rec_cfg["end_x"] = (end_x // bin_x) - ((nx % bin_x) != 0)
            rec_cfg["end_y"] = (end_y // bin_y) - ((ny % bin_y) != 0)
            rec_cfg["end_z"] = (end_z // bin_z) - ((nz % bin_z) != 0)
            rec_cfg["start_x"] = (rec_cfg["start_x"] // bin_x)
            rec_cfg["start_y"] = (rec_cfg["start_y"] // bin_y)
            rec_cfg["start_z"] = (rec_cfg["start_z"] // bin_z)
            # Update the Rotation center
            # TODO axis_corrections
            rot_c = self.dataset_infos.axis_position
            nx0, nz0 = self.dataset_infos._radio_dims_notbinned
            bin_fact = nx0 / nx
            if rot_c is not None: # user-specified
                rot_c /= bin_fact # float
            else:
                rot_c = (nx - 1)/2.
            self.dataset_infos.axis_position = rot_c
            self.nabu_config["reconstruction"]["rotation_axis_position"] = rot_c


    def check_output_file(self):
        out_cfg = self.nabu_config["output"]
        out_fname = os.path.join(out_cfg["location"], out_cfg["file_prefix"] + out_cfg["file_format"])
        if os.path.exists(out_fname):
            raise ValueError("File %s already exists" % out_fname)


    def handle_processing_mode(self):
        mode = self.nabu_config["resources"]["method"]
        if mode == "preview":
            print("Warning: the method 'preview' was selected. This means that the data volume will be binned so that everything fits in memory.")
            # TODO automatically compute binning/subsampling factors as a function of lowest memory (GPU)
            self.nabu_config["dataset"]["binning"] = 2
            self.nabu_config["dataset"]["binning_z"] = 2
            self.nabu_config["dataset"]["projections_subsampling"] = 2
        # TODO handle other modes





from os import path, mkdir
from ..resources.logger import LoggerOrPrint
from ..resources.utils import is_hdf5_extension
from ..utils import check_supported
from ..io.writer import Writers

#
# Decorators and callback mechanism
#

def use_options(step_name, step_attr):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if step_name not in self.processing_steps:
                self.__setattr__(step_attr, None)
                return
            self._steps_name2component[step_name] = step_attr
            self._steps_component2name[step_attr] = step_name
            return func(*args, **kwargs)
        return wrapper
    return decorator


def pipeline_step(step_attr, step_desc):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if self.__getattribute__(step_attr) is None:
                return
            self.logger.info(step_desc)
            res = func(*args, **kwargs)
            self.logger.debug("End " + step_desc)
            callback = self._callbacks.get(self._steps_component2name[step_attr], None)
            if callback is not None:
                callback(self)
            return res
        return wrapper
    return decorator


#
# Writer
#

class WriterConfigurator:
    def __init__(self, options, output_dir=None, start_index=None, logger=None, nx_info=None):
        """
        Create a Writer from a set of parameters.

        Parameters
        ----------
        options: dict
            A dictionary of options for writer. Must have at least "file_format", "location"
            and "file_prefix" keys.
        output_dir: str, optional
            Directory where the file(s) will be written. If not provided, it will
            be set to options["location"].
        start_index: int, optional
            Index to start the files numbering (filename_0123.ext).
            Default is 0.
            Ignored for HDF5 extension.
        logger: nabu.resources.logger.Logger, optional
            Logger object
        nx_info: dict, optional
            Dictionary containing the nexus information.
        """
        self.logger = LoggerOrPrint(logger)
        self.options = options
        self.start_index = start_index

        file_format = options["file_format"]
        check_supported(file_format, list(Writers.keys()), "output file format")

        self._set_output_dir(output_dir)
        self._set_file_name()

        # Init Writer
        writer_cls = Writers[file_format]
        writer_args = [self.fname]
        writer_kwargs = {}
        self._writer_exec_args = []
        self._writer_exec_kwargs = {}

        if is_hdf5_extension(file_format):
            writer_kwargs["entry"] = nx_info["entry"]
            writer_kwargs["filemode"] = "a"
            writer_kwargs["overwrite"] = self.options["overwrite"]
            self._writer_exec_args.append(nx_info["process_name"])
            self._writer_exec_kwargs["processing_index"] = nx_info["processing_index"]
            self._writer_exec_kwargs["config"] = nx_info["config"]
        else:
            writer_kwargs["start_index"] = self.start_index
        self.writer = writer_cls(*writer_args, **writer_kwargs)


    def _set_output_dir(self, output_dir):
        if output_dir is None:
            output_dir = path.join(
                self.options["location"],
                self._orig_file_prefix
            )
        self.output_dir = output_dir
        if path.exists(self.output_dir):
            if not(path.isdir(self.output_dir)):
                raise ValueError(
                    "Unable to create directory %s: already exists and is not a directory"
                    % self.output_dir
                )
        else:
            self.logger.debug("Creating directory %s" % self.output_dir)
            mkdir(self.output_dir)


    def _set_file_name(self):
        self.fname = path.join(
            self.output_dir,
            self.options["file_prefix"] + "." + self.options["file_format"]
        )
        if path.exists(self.fname):
            err = "File already exists: %s" % self.fname
            if self.options["overwrite"]:
                if self.options.get("warn_overwrite", True):
                    self.logger.warning(err + ". It will be overwritten as requested in configuration")
                    self.options["warn_overwrite"] = False
            else:
                self.logger.fatal(err)
                raise ValueError(err)



#!/usr/bin/env python
# coding: utf-8


from setuptools import setup
import os
# Relative-first
from nabu import version, __nabu_modules__

def setup_package():
    packages_folders = __nabu_modules__
    packages = ["nabu"]
    package_dir = {"nabu": "nabu"}
    for f in packages_folders:
        modulename = str("nabu.%s" % f)
        packages.append(modulename)
        package_dir[modulename] = os.path.join("nabu", f)

        module_test_dirname = os.path.join(package_dir[modulename], "tests")
        if os.path.isdir(module_test_dirname):
            modulename_test = str("%s.tests" % modulename)
            packages.append(modulename_test)
            package_dir[modulename_test] = module_test_dirname

    other_modules = ["nabu.resources.cli"]
    for mod in other_modules:
        packages.append(mod)
        package_dir[mod] = os.path.join(*(mod.split(".")))

    doc_requires = [
        'sphinx',
        'cloud_sptheme',
        'recommonmark',
        'nbsphinx',
    ]

    setup(
        name='nabu',
        author='Pierre Paleo',
        version=version,
        author_email = "pierre.paleo@esrf.fr",
        maintainer = "Pierre Paleo",
        maintainer_email = "pierre.paleo@esrf.fr",

        packages=packages,
        package_dir = package_dir,
        package_data = {
            'nabu.cuda': [
                'src/*.cu',
                'src/*.h',
            ],
            'nabu.resources': [
                'templates/*.ini',
            ],
        },
        install_requires = [
            'psutil',
            'pytest',
            'numpy > 1.9.0',
            'silx >= 0.12.0',
            'distributed', # >= 2.10.0',
            'dask_jobqueue',
            # ~ 'pycuda',
            # ~ 'pyopencl',
            'tomoscan >= 0.3.2',
        ],
        long_description = """
        Nabu - Tomography software
        """,

        entry_points = {
            'console_scripts': [
                "nabu-test=nabu.tests:nabu_test",
                "nabu-config=nabu.resources.cli.bootstrap:bootstrap",
                "nabu-zsplit=nabu.resources.cli.nx_z_splitter:zsplit",
                "nabu-histogram=nabu.resources.cli.histogram:histogram_cli",
                "nabu=nabu.resources.cli.reconstruct:main",
            ],
        },

        zip_safe=True
    )


if __name__ == "__main__":
    setup_package()



import numpy as np
from math import sqrt, pi

from ..utils import updiv, get_cuda_srcfile, _sizeof, nextpow2, convert_index
from ..cuda.utils import copy_array
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel
from .filtering import SinoFilter
import pycuda.driver as cuda
from pycuda import gpuarray as garray


class Backprojector(CudaProcessing):
    """
    Cuda Backprojector.
    """
    def __init__(
        self,
        sino_shape,
        slice_shape=None,
        angles=None,
        rot_center=None,
        filter_name=None,
        slice_roi=None,
        scale_factor=None,
        extra_options={},
        cuda_options={},
    ):
        """
        Initialize a Cuda Backprojector.

        Parameters
        -----------
        sino_shape: tuple
            Shape of the sinogram, in the form `(n_angles, detector_width)`
            (for backprojecting one sinogram) or `(n_sinos, n_angles, detector_width)`.
        slice_shape: int or tuple, optional
            Shape of the slice. By default, the slice shape is (n_x, n_x) where
            `n_x = detector_width`
        angles: array-like, optional
            Rotation anles in radians.
            By default, angles are equispaced between [0, pi[.
        rot_center: float, optional
            Rotation axis position. Default is `(detector_width - 1)/2.0`
        filter_name: str, optional
            Name of the filter for filtered-backprojection.
        slice_roi: tuple, optional.
            Whether to backproject in a restricted area.
            If set, it must be in the form (start_x, end_x, start_y, end_y).
            `end_x` and `end_y` are non inclusive ! For example if the detector has
            2048 pixels horizontally, then you can choose `start_x=0` and `end_x=2048`.
            If one of the value is set to None, it is replaced with a default value
            (0 for start, n_x and n_y for end)
        scale_factor: float, optional
            Scaling factor for backprojection.
            For example, to get the linear absorption coefficient in 1/cm,
            this factor has to be set as the pixel size in cm.
        extra_options: dict, optional
            Advanced extra options.
             See the "Extra options" section for more information.
        cuda_options: dict, optional
            Cuda options passed to the CudaProcessing class.

        Extra options
        --------------
        The parameter `extra_options` is a dictionary with the following defaults:
           - "padding_mode": "zeros"
             Padding mode when filtering the sinogram. Can be "zeros" or "edges".
           - "axis_correction": None
             Whether to set a correction for the rotation axis.
             If set, this should be an array with as many elements as the number
             of angles. This is useful when there is an horizontal displacement
             of the rotation axis.
        """
        super().__init__(**cuda_options)
        self._configure_extra_options(scale_factor, extra_options=extra_options)
        self._init_geometry(sino_shape, slice_shape, angles, rot_center, slice_roi)
        self._init_filter(filter_name)
        self._allocate_memory()
        self._compute_angles()
        self._compile_kernels()
        self._bind_textures()


    def _configure_extra_options(self, scale_factor, extra_options={}):
        if scale_factor is None:
            scale_factor = 1.
        self._backproj_scale_factor = scale_factor
        self._axis_array = None
        self.extra_options = {
            "padding_mode": "zeros",
            "axis_correction": None,
        }
        self.extra_options.update(extra_options)
        self._axis_array = self.extra_options["axis_correction"]


    def _init_geometry(self, sino_shape, slice_shape, angles, rot_center, slice_roi):
        if slice_shape is not None and slice_roi is not None:
            raise ValueError("slice_shape and slice_roi cannot be used together")
        self.sino_shape = sino_shape
        if len(sino_shape) == 2:
            n_angles, dwidth = sino_shape
            n_slices = 1
        elif len(sino_shape) == 3:
            n_slices, n_angles, dwidth = sino_shape
        else:
            raise ValueError("Expected 2D or 3D sinogram")
        n_sinos = n_slices
        self.dwidth = dwidth
        self.n_slices = n_slices
        self._set_slice_shape(slice_shape)
        self.n_sinos = n_sinos
        self.rot_center = rot_center or (self.dwidth - 1)/2.
        self.axis_pos = self.rot_center
        self._set_angles(angles, n_angles)
        self._set_slice_roi(slice_roi)
        self._set_axis_corr()


    def _set_slice_shape(self, slice_shape):
        n_y = self.dwidth
        n_x = self.dwidth
        if slice_shape is not None:
            if np.isscalar(slice_shape):
                slice_shape = (slice_shape, slice_shape)
            n_y, n_x = slice_shape
        self.n_x = n_x
        self.n_y = n_y
        self.slice_shape = (n_y, n_x)
        if self.n_slices > 1:
            self.slice_shape = (self.n_slices,) + self.slice_shape


    def _set_angles(self, angles, n_angles):
        self.n_angles = n_angles
        if angles is None:
            angles = n_angles
        if np.isscalar(angles):
            angles = np.linspace(0, np.pi, angles, False)
        else:
            assert len(angles) == self.n_angles
        self.angles = angles


    def _set_slice_roi(self, slice_roi):
        self.offsets = {"x": 0, "y": 0}
        self.slice_roi = slice_roi
        if slice_roi is None:
            return
        start_x, end_x, start_y, end_y = slice_roi
        # convert negative indices
        dwidth = self.dwidth
        start_x = convert_index(start_x, dwidth, 0)
        start_y = convert_index(start_y, dwidth, 0)
        end_x = convert_index(end_x, dwidth, dwidth)
        end_y = convert_index(end_y, dwidth, dwidth)
        self.slice_shape = (end_y - start_y, end_x - start_x)
        if self.n_slices > 1:
            self.slice_shape = (n_slices,) + self.slice_shape
        self.n_x = self.slice_shape[-1]
        self.n_y = self.slice_shape[-2]
        self.offsets = {"x": start_x, "y": start_y}


    def _allocate_memory(self):
        self._d_sino_cua = cuda.np_to_array(np.zeros(self.sino_shape, "f"), "C")
        # 1D textures are not supported in pycuda
        self.h_msin = np.zeros((1, self.n_angles), "f")
        self.h_cos = np.zeros((1, self.n_angles), "f")
        self._d_sino = garray.zeros(self.sino_shape, "f")
        self._init_arrays_to_none(["_d_slice"])


    def _compute_angles(self):
        self.h_cos[0] = np.cos(self.angles).astype("f")
        self.h_msin[0] = (-np.sin(self.angles)).astype("f")
        self._d_msin = garray.to_gpu(self.h_msin[0])
        self._d_cos = garray.to_gpu(self.h_cos[0])
        if self._axis_correction is not None:
            self._d_axcorr = garray.to_gpu(self._axis_correction)


    def _set_axis_corr(self):
        axcorr = self.extra_options["axis_correction"]
        self._axis_correction = axcorr
        if axcorr is None:
            return
        assert len(axcorr) == self.n_angles
        self._axis_correction = np.zeros((1, self.n_angles), dtype=np.float32)
        self._axis_correction[0, :] = axcorr[:]


    def _init_filter(self, filter_name):
        self.filter_name = filter_name
        self.sino_filter = SinoFilter(
            self.sino_shape,
            filter_name=self.filter_name,
            padding_mode=self.extra_options["padding_mode"],
            ctx=self.ctx,
        )


    def _get_kernel_signature(self):
        kern_full_sig = list("PiiifiiiiPPPf")
        if self.n_sinos == 1:
            kern_full_sig[3] = ""
        if self._axis_correction is None:
            kern_full_sig[11] = ""
        return "".join(kern_full_sig)


    def _get_kernel_options(self):
        tex_name = "tex_projections"
        sourcemodule_options = []
        # We use blocks of 16*16 (see why in kernel doc), and one thread
        # handles 2 pixels per dimension.
        block = (16, 16, 1)
        # The Cuda kernel is optimized for 16x16 threads blocks
        # If one of the dimensions is smaller than 16, it has to be addapted
        if self.n_x < 16 or self.n_y < 16:
            tpb_x = min(int(nextpow2(self.n_x)), 16)
            tpb_y = min(int(nextpow2(self.n_y)), 16)
            block = (tpb_x, tpb_y, 1)
            sourcemodule_options.append("-DSHARED_SIZE=%d" % (tpb_x * tpb_y))
        grid = (
            updiv(updiv(self.n_x, block[0]), 2),
            updiv(updiv(self.n_y, block[1]), 2)
        )
        if self.n_sinos > 1:
            tex_name = "tex_projections3D"
            sourcemodule_options.append("-DBACKPROJ3D")
            grid += (self.n_sinos,)
        shared_size = int(np.prod(block)) * 2
        if self._axis_correction is not None:
            sourcemodule_options.append("-DDO_AXIS_CORRECTION")
            shared_size += int(np.prod(block))
        shared_size *= 4 # sizeof(float32)
        self._kernel_options = {
            "file_name": get_cuda_srcfile("backproj.cu"),
            "kernel_name": "backproj",
            "kernel_signature": self._get_kernel_signature(),
            "texture_name": tex_name,
            "sourcemodule_options": sourcemodule_options,
            "grid": grid,
            "block": block,
            "shared_size": shared_size
        }


    def _compile_kernels(self):
        self._get_kernel_options()
        kern_opts = self._kernel_options
        # Configure backprojector
        self.gpu_projector = CudaKernel(
            kern_opts["kernel_name"],
            filename=kern_opts["file_name"],
            options=kern_opts["sourcemodule_options"]
        )
        self.texref_proj = self.gpu_projector.module.get_texref(
            kern_opts["texture_name"]
        )
        self.texref_proj.set_filter_mode(cuda.filter_mode.LINEAR)
        self.gpu_projector.prepare(kern_opts["kernel_signature"], [self.texref_proj])
        # Prepare kernel arguments
        self.kern_proj_args = [
            None, # output d_slice holder
            self.n_angles,
            self.dwidth,
            self.axis_pos,
            self.n_x,
            self.n_y,
            self.offsets["x"],
            self.offsets["y"],
            self._d_cos,
            self._d_msin,
            self._backproj_scale_factor
        ]
        if self.n_sinos > 1:
            self.kern_proj_args.insert(3, self.n_slices)
        if self._axis_correction is not None:
            self.kern_proj_args.insert(-1, self._d_axcorr)
        self.kern_proj_kwargs = {
            "grid": kern_opts["grid"],
            "block": kern_opts["block"],
            "shared_size": kern_opts["shared_size"],
        }


    def _bind_textures(self):
        self.texref_proj.set_array(self._d_sino_cua)


    def _set_output(self, output, check=False):
        if output is None:
            self._allocate_array("_d_slice", self.slice_shape, dtype=np.float32)
            return self._d_slice
        if check:
            assert output.dtype == np.float32
            assert output.shape == self.slice_shape, "Expected output shape %s but got %s" % (self.slice_shape, output.shape)
        if isinstance(output, garray.GPUArray):
            return output.gpudata
        else: # pycuda.driver.DeviceAllocation ?
            return output


    def backproj(self, sino, output=None, do_checks=True):
        copy_array(self._d_sino_cua, sino, check=do_checks)
        d_slice = self._set_output(output, check=do_checks)
        self.kern_proj_args[0] = d_slice
        self.gpu_projector(
            *self.kern_proj_args,
            **self.kern_proj_kwargs
        )
        if output is not None:
            return output
        else:
            return self._d_slice.get()


    def filtered_backprojection(self, sino, output=None):
        self.sino_filter(sino, output=self._d_sino)
        return self.backproj(self._d_sino, output=output)


    fbp = filtered_backprojection # shorthand



#!/usr/bin/env python

"""
Validate: tool for validating a configuration file.
This tool will inspect each entry of the configuration file and ensure that:

   - The dataset can be accessed (and the output location is writeable)
   - The relevant metadata can be accessed (pixel size, energy, angles, ...)
   - Each value of the configuration file is correct (eg. invalid values)
"""

import os
from .utils import parse_params_values
from .cli_configs import ValidateConfig
from ...io.config import NabuConfigParser, validate_nabu_config
from ..dataset_analyzer import analyze_dataset
from ..dataset_validator import NabuValidator


def get_config(fname):
    if not(os.path.isfile(fname)):
        print("Error: file %s not found" % fname)
        exit(-1)
    parser = NabuConfigParser(fname)
    config = parser.conf_dict
    return config


# Obsolete with ProcessConfig
def validate_conffile(fname):
    # Read the configuration file and extract the key/values
    config = get_config(fname)

    # Validation step 1: check the config file alone
    config = validate_nabu_config(config)

    # Browse the dataset
    dataset_structure = analyze_dataset(config["dataset"]["location"])

    # Validation step 2: check the consistency of dataset and config file
    validator = NabuValidator(config, dataset_structure)
    validator.perform_all_checks()

    # Remove unused radios (modifies dataset_structure)
    validator.remove_unused_radios()

    return validator


# Obsolete with ProcessConfig
def validate_conffile_noexcept(fname, print_ok=True):
    try:
        validator = validate_conffile(fname)
    except Exception as exc:
        print("Error while checking %s:" % fname)
        print(exc)
        exit(-1)
    if print_ok:
        print("Configuration file %s is valid" % fname)
    return validator


def main():
    args = parse_params_values(
        ValidateConfig,
        parser_description="Validate a Nabu configuration file."
    )
    fname = args["input_file"]
    validate_conffile_noexcept(fname)

if __name__ == "__main__":
    main()



from os.path import join, isfile, basename
from math import ceil
import gc
from psutil import virtual_memory
from silx.io import get_data
from silx.io.url import DataUrl
from ..resources.logger import LoggerOrPrint
from ..io.writer import merge_hdf5_files, NXProcessWriter
from ..cuda.utils import collect_cuda_gpus
from ..resources.computations import estimate_chunk_size
from ..preproc.phase import compute_paganin_margin
from ..app.fullfield_cuda import CudaFullFieldPipeline, CudaFullFieldPipelineLimitedMemory
from ..misc.histogram import PartialHistogram, add_last_bin, hist_as_2Darray

def get_gpus_ids(resources_cfg):
    gpus_ids = resources_cfg["gpu_id"]
    if gpus_ids != []:
        return gpus_ids
    # TODO (?) heuristic to pick the best gpus
    return list(range(resources_cfg["gpus"]))



class LocalReconstruction:

    _pipeline_cls = CudaFullFieldPipeline

    def __init__(self, process_config, logger=None, extra_options=None, cuda_options=None):
        self.process_config = process_config
        self.logger = LoggerOrPrint(logger)
        self._set_extra_options(extra_options)
        self._get_reconstruction_range()
        self._get_resources()
        self._compute_max_chunk_size()
        self._compute_phase_margin()
        self._get_pipeline_class()
        self._compute_volume_chunks()
        self.pipeline = None
        self.cuda_options = cuda_options


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "gpu_mem_fraction": 0.9,
            "cpu_mem_fraction": 0.9,
            "use_phase_margin": True,
            "max_chunk_size": None,
            "phase_margin": None,
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options
        self.gpu_mem_fraction = self.extra_options["gpu_mem_fraction"]
        self.cpu_mem_fraction = self.extra_options["cpu_mem_fraction"]
        self.use_phase_margin = self.extra_options["use_phase_margin"]


    def _get_reconstruction_range(self):
        rec_cfg = self.process_config.nabu_config["reconstruction"]
        self.z_min = rec_cfg["start_z"]
        self.z_max = rec_cfg["end_z"] + 1
        self.delta_z = self.z_max - self.z_min


    def _get_resources(self):
        self.resources = {}
        self._get_gpu()
        self._get_memory()


    def _get_memory(self):
        vm = virtual_memory()
        self.resources["mem_avail_GB"] = vm.available / 1e9


    def _get_gpu(self):
        gpus = get_gpus_ids(self.process_config.nabu_config["resources"])
        if len(gpus) == 0:
            raise ValueError("Need at least one GPU")
        if len(gpus) > 1:
            raise ValueError("This class does not support more than one GPU")
        self.resources["gpu_id"] = self._gpu_id = gpus[0]
        self.resources["gpus"] = collect_cuda_gpus()


    def _compute_max_chunk_size(self):
        gpu_mem = self.resources["gpus"][self._gpu_id]["memory_GB"] * self.gpu_mem_fraction
        cpu_mem = self.resources["mem_avail_GB"] * self.cpu_mem_fraction
        self.gpu_max_chunk_size = estimate_chunk_size(
            gpu_mem,
            self.process_config,
            chunk_step=10
        )
        self.cpu_max_chunk_size = estimate_chunk_size(
            cpu_mem,
            self.process_config,
            chunk_step=10
        )
        user_max_chunk_size = self.extra_options["max_chunk_size"]
        if user_max_chunk_size is not None:
            self.gpu_max_chunk_size = min(self.gpu_max_chunk_size, user_max_chunk_size)
            self.cpu_max_chunk_size = min(self.cpu_max_chunk_size, user_max_chunk_size)


    def _compute_phase_margin(self):
        if "phase" not in self.process_config.processing_steps:
            self._phase_margin = (0, 0)
            self._margin_v = self._phase_margin[0]
            return
        radio_shape = self.process_config.dataset_infos.radio_dims[::-1]
        opts = self.process_config.processing_options["phase"]
        user_phase_margin = self.extra_options["phase_margin"]
        if user_phase_margin is not None and user_phase_margin > 0:
            margin_v, margin_h = user_phase_margin, user_phase_margin
        else:
            margin_v, margin_h = compute_paganin_margin(
                radio_shape,
                distance=opts["distance_cm"],
                energy=opts["energy_kev"],
                delta_beta=opts["delta_beta"],
                pixel_size=opts["pixel_size_microns"],
                padding=opts["padding_type"]
            )
        self._phase_margin = (margin_v, margin_h)
        self._margin_v = self._phase_margin[0]
        self.logger.info("Estimated phase margin: %d pixels" % self._margin_v)


    def _get_pipeline_class(self):
        self._limited_mem = False
        # Actually less in some cases (margin_far_up + margin_far_down instead of 2*margin_v).
        # But we want to use only one class for all stages.
        chunk_size_for_one_slice = 1 + 2 * self._margin_v
        chunk_is_too_small = False
        if chunk_size_for_one_slice > self.gpu_max_chunk_size:
            msg = str(
                "Phase margin is %d, so we need to process at least %d detector rows. However, the available memory enables to process only %d rows at once"
                % (self._margin_v, chunk_size_for_one_slice, self.gpu_max_chunk_size)
            )
            chunk_is_too_small = True
        if self._margin_v > self.gpu_max_chunk_size//3:
            n_slices = max(1, self.gpu_max_chunk_size - (2 * self._margin_v))
            n_stages = ceil(self.delta_z / n_slices)
            if n_stages > 1:
                # In this case, using CudaFlatField + margin would lead to many stages
                msg = str(
                    "Phase margin (%d) is too big for chunk size (%d)"
                    % (self._margin_v, self.gpu_max_chunk_size)
                )
                chunk_is_too_small = True
        if chunk_is_too_small:
            self.logger.warning(msg)
            if self.use_phase_margin:
                self._pipeline_cls = CudaFullFieldPipelineLimitedMemory
                self.logger.warning("Using CudaFullFieldPipelineLimitedMemory")
                self._limited_mem = True
            else:
                self._phase_margin = (0, 0)
                self._margin_v = self._phase_margin[0]
                self._pipeline_cls = CudaFullFieldPipeline
                self.logger.warning("Using CudaFullFieldPipeline without margin")


    def _compute_volume_chunks(self):
        n_z = self.process_config.dataset_infos._radio_dims_notbinned[1]
        margin_v = self._margin_v
        self._margin_far_up = min(margin_v, self.z_min)
        self._margin_far_down = min(margin_v, n_z - (self.z_max + 1))
        # | margin_up |     n_slices    |  margin_down |
        # |-----------|-----------------|--------------|
        # |----------------------------------------------------|
        #                    delta_z
        if self._limited_mem:
            n_slices = self.cpu_max_chunk_size
        else:
            n_slices = self.gpu_max_chunk_size - (2 * margin_v)
        tasks = []
        if n_slices >= self.delta_z:
            # In this case we can do everything in a single stage
            n_slices = self.delta_z
            (margin_up, margin_down) = (self._margin_far_up, self._margin_far_down)
            tasks.append({
                "sub_region": (self.z_min - margin_up, self.z_max + margin_down),
                "phase_margin": ((margin_up, margin_down), (0, 0))
            })
        if self.gpu_max_chunk_size >= self.delta_z and self.z_min == 0 and self.z_max == n_z:
            # In this case we can do everything in a single stage, without margin
            n_slices = self.delta_z
            tasks.append({
                "sub_region": (self.z_min, self.z_max),
                "phase_margin": None,
            })
        else:
            # In this case there are at least two stages
            n_stages = ceil(self.delta_z / n_slices)
            tasks = []
            curr_z_min = self.z_min
            curr_z_max = self.z_min + n_slices
            for i in range(n_stages):
                margin_up = min(margin_v, curr_z_min)
                margin_down = min(margin_v, max(n_z - curr_z_max, 0))
                if curr_z_max + margin_down >= self.z_max:
                    curr_z_max -= (curr_z_max - (self.z_max + 0))
                    margin_down = min(margin_v, max(n_z - 1 - curr_z_max, 0))
                tasks.append({
                    "sub_region": (curr_z_min - margin_up, curr_z_max + margin_down),
                    "phase_margin": ((margin_up, margin_down), (0, 0))
                })
                if curr_z_max == self.z_max:
                    # No need for further tasks
                    break
                curr_z_min += n_slices
                curr_z_max += n_slices
        self.tasks = tasks
        self.n_slices = n_slices


    def _print_tasks(self):
        for task in self.tasks:
            margin_up, margin_down = task["phase_margin"][0]
            s_u, s_d = task["sub_region"]
            print(
                "Top Margin: [%04d, %04d[  |  Slices: [%04d, %04d[  |  Bottom Margin: [%04d, %04d["
                % (
                    s_u, s_u + margin_up,
                    s_u + margin_up, s_d - margin_down,
                    s_d - margin_down, s_d
                )
            )


    def _instantiate_pipeline(self, task):
        self.logger.debug("Creating a new pipeline object")
        args = [self.process_config, task["sub_region"]]
        if self._limited_mem:
            # Adapt chunk size so that [margin_up, chunk_size, margin_down]
            # is equal to delta_z.
            chunk_size = self.gpu_max_chunk_size
            dz = self._get_delta_z(task)
            margin_v_tot = sum(task["phase_margin"][0])
            args.append(chunk_size)
        pipeline = self._pipeline_cls(
            *args,
            logger=self.logger,
            phase_margin=task["phase_margin"],
            cuda_options=self.cuda_options
        )
        self.pipeline = pipeline


    def _instantiate_pipeline_if_necessary(self, current_task, other_task):
        """
        Instantiate a pipeline only if current_task has a different "delta z" than other_task
        """
        if self.pipeline is None:
            self._instantiate_pipeline(current_task)
            return
        dz_cur = self._get_delta_z(current_task)
        dz_other = self._get_delta_z(other_task)
        if dz_cur != dz_other:
            self.logger.debug("Destroying pipeline instance and releasing memory")
            self.pipeline = None
            # Not elegant, but for now the only way to release Cuda memory
            gc.collect()
            self._instantiate_pipeline(current_task)


    @staticmethod
    def _get_delta_z(task):
        # will have to be modified if sub_region accounts for x-subregion
        return task["sub_region"][1] - task["sub_region"][0]


    def reconstruct(self):
        tasks = self.tasks
        self.results = {}
        prev_task = tasks[0]
        for task in tasks:
            self.logger.info("\nProcessing sub-volume %s" % (str(task["sub_region"])))
            self._instantiate_pipeline_if_necessary(task, prev_task)
            self.pipeline.process_chunk(sub_region=task["sub_region"])
            self.results[self.pipeline.sub_region[-2:]] = self.pipeline.writer.fname
            prev_task = task


    def merge_hdf5_reconstructions(self, output_file=None):
        """
        Merge existing reconstructions by creating a HDF5 virtual dataset.

        Parameters
        ----------
        output_file: str, optional
            Output file name. If not given, the file prefix in section "output"
            of nabu config will be taken.
        """
        out_cfg = self.process_config.nabu_config["output"]
        if output_file is None:
            output_file = join(out_cfg["location"], out_cfg["file_prefix"]) + ".hdf5"
        if isfile(output_file):
            msg = str("File %s already exists" % output_file)
            if out_cfg["overwrite_results"]:
                msg += ". Overwriting as requested in configuration file"
                self.logger.warning(msg)
            else:
                msg += ". Set overwrite_results to True in [output] to overwrite existing files."
                self.logger.fatal(msg)
                raise ValueError(msg)
        files = list(self.results.values())
        if files == []:
            self.logger.error("No files to merge")
            return
        files.sort()
        self._local_files = [
            join(out_cfg["file_prefix"], basename(fname))
            for fname in files
        ]
        entry = getattr(self.process_config.dataset_infos.dataset_scanner, "entry", "entry")
        # TODO "reconstruction" is hard-coded, might be problematic in the future
        h5_path = join(entry, *["reconstruction", "results", "data"])
        process_name = "reconstruction"
        #
        self.logger.info("Merging reconstructions to %s" % output_file)
        merge_hdf5_files(
            self._local_files, h5_path, output_file, process_name,
            output_entry=entry,
            output_filemode="a",
            processing_index=0,
            config={
                "reconstruction_stages": {
                    str(k): v for k, v in zip(self.results.keys(), self._local_files)
                },
                "nabu_config": self.process_config.nabu_config
            },
            base_dir=out_cfg["location"],
            overwrite=out_cfg["overwrite_results"]
        )
        # TODO dont hard-code process name
        self.merge_histograms(entry, "histogram", output_file, entry, process_name)


    def merge_histograms(
        self, masterfile_entry, masterfile_process_name,
        output_file, output_entry, output_process_name
    ):
        """
        Merge the partial histograms
        """
        if not(self.process_config.nabu_config["postproc"]["output_histogram"]):
            return
        self.logger.info("Merging histograms")

        #
        h5_path = join(masterfile_entry, *[masterfile_process_name, "results", "data"])
        #
        files = sorted(self.results.values())
        data_urls = []
        for fname in files:
            url = DataUrl(
                file_path=fname, data_path=h5_path, data_slice=None, scheme="silx"
            )
            data_urls.append(url)
        histograms = []
        for data_url in data_urls:
            h2D = get_data(data_url)
            histograms.append(
                (h2D[0], add_last_bin(h2D[1]))
            )
        histograms_merger = PartialHistogram( # TODO configurable
            method="fixed_bins_number", num_bins=histograms[0][0].size
        )
        merged_hist = histograms_merger.merge_histograms(histograms)

        writer = NXProcessWriter(
            output_file, entry=output_entry, filemode="a", overwrite=True
        )
        writer.write(
            hist_as_2Darray(merged_hist),
            "histogram", # TODO don't hard-code
            processing_index=1,
            config={
                "files": self._local_files,
                "bins": self.process_config.nabu_config["postproc"]["histogram_bins"],
            }
        )




#!/usr/bin/env python

from ... import version
from .utils import parse_params_values
from ..utils import is_hdf5_extension
from .cli_configs import ReconstructConfig

def get_subregion(slices_indices, radio_nz):
    if len(slices_indices) == 0:
        return (0, radio_nz-1)
    if slices_indices == "all":
        return (0, radio_nz-1)
    if slices_indices == "first":
        return (0, 0)
    if slices_indices == "middle":
        return (radio_nz // 2, radio_nz // 2)
    if slices_indices == "last":
        return (radio_nz-1, radio_nz-1)
    try:
        if "-" in slices_indices:
            z_start, z_stop = slices_indices.split("-")
            z_start = int(z_start)
            z_stop = int(z_stop)
        else:
            z_idx = int(slices_indices)
            z_start = z_idx
            z_stop = z_idx
    except Exception as exc:
        print("Could not interpret slice indices: %s")
        print(exc)
        exit(1)
    return (z_start, z_stop)


def get_reconstruction_region(arg_slice, process_config, logger):
    n_z = process_config.dataset_infos._radio_dims_notbinned[-1]
    rec_opts = process_config.nabu_config["reconstruction"]

    if arg_slice == "":
        subregion = (rec_opts["start_z"], rec_opts["end_z"])
    else:
        # Overwrite config "start_z" and "end_z". Should this be deprecated ?
        subregion = get_subregion(
            arg_slice,
            process_config.dataset_infos.radio_dims[-1]
        )
        start_z, end_z = subregion
        logger.info(
            "Overwriting config file start_z to %d and end_z to %d"
            % (start_z, end_z)
        )
        process_config.nabu_config["reconstruction"]["start_z"] = start_z
        process_config.nabu_config["reconstruction"]["end_z"] = end_z

    delta_z = subregion[-1] - subregion[-2]

    _, radio_nz = process_config.dataset_infos.radio_dims
    if subregion[0] >= radio_nz or subregion[1] >= radio_nz:
        msg = str(
            "Reconstruction outside range: asked [zmin, zmax] = [%d, %d] but detector vertical dimension is %d"
            % (subregion[0], subregion[1], radio_nz)
        )
        logger.fatal(msg)
        raise ValueError(msg)

    return subregion


def get_log_file(arg_logfile, legacy_arg_logfile, forbidden=None):
    default_arg_val = "nabu.log"
    # Compat. log_file --> logfile
    if legacy_arg_logfile != default_arg_val:
        logfile = legacy_arg_logfile
    else:
        logfile = arg_logfile
    #
    if forbidden is None:
        forbidden = []
    for forbidden_val in forbidden:
        if logfile == forbidden_val:
            print("Error: --logfile argument cannot have the value %s" % forbidden_val)
            exit(1)
    return logfile


def main():
    args = parse_params_values(
        ReconstructConfig,
        parser_description="Perform a tomographic reconstruction.",
        program_version="nabu " + version
    )

    # Imports are done here, otherwise "nabu --version" takes forever
    from ...resources.processconfig import ProcessConfig
    from ...cuda.utils import __has_pycuda__
    if __has_pycuda__:
        from ...app.local_reconstruction import LocalReconstruction
    else:
        print("Error: need cuda and pycuda for reconstruction")
        exit(1)
    from ..logger import Logger
    #

    # A crash with scikit-cuda happens only on PPC64 platform if and nvidia-persistenced is running.
    # On such machines, a warm-up has to be done.
    import platform
    if platform.machine() == "ppc64le":
        from silx.math.fft.cufft import CUFFT
    #

    proc = ProcessConfig(args["input_file"])
    logfile = get_log_file(args["logfile"], args["log_file"], forbidden=[args["input_file"]])

    logger = Logger(
        "nabu",
        level=proc.nabu_config["about"]["verbosity"],
        logfile=logfile
    )

    subregion = get_reconstruction_region(args["slice"].strip(), proc, logger)
    logger.info("Going to reconstruct slices %s" % str(subregion))
    subregion = (None, None) + subregion

    # (hopefully) temporary patch
    if "phase" in proc.processing_steps:
        if args["energy"] > 0:
            logger.warning("Using user-provided energy %.2f keV" % args["energy"])
            proc.dataset_infos.dataset_scanner._energy = args["energy"]
            proc.processing_options["phase"]["energy_kev"] = args["energy"]
        if proc.dataset_infos.energy  < 1e-3 and proc.nabu_config["phase"]["method"] != None:
            msg = "No information on energy. Cannot retrieve phase. Please use the --energy option"
            logger.fatal(msg)
            raise ValueError(msg)
    #

    R = LocalReconstruction(
        proc,
        logger=logger,
        extra_options={
            "gpu_mem_fraction": args["gpu_mem_fraction"],
            "cpu_mem_fraction": args["cpu_mem_fraction"],
            "use_phase_margin": args["use_phase_margin"],
            "max_chunk_size": args["max_chunk_size"] if args["max_chunk_size"] > 0 else None,
            "phase_margin": args["phase_margin"],
        }
    )

    R.reconstruct()
    if is_hdf5_extension(proc.nabu_config["output"]["file_format"]):
        R.merge_hdf5_reconstructions()




if __name__ == "__main__":
    main()
